from pandas import DataFrame

from metricflow.constraints.time_constraint import TimeRangeConstraint
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.sql_client import SqlClient
from metricflow.time.time_constants import ISO8601_PYTHON_FORMAT, ISO8601_PYTHON_TS_FORMAT


def test_date_spine_date_range(sql_client: SqlClient, time_spine_source: TimeSpineSource) -> None:  # noqa: D
    range_df: DataFrame = sql_client.query(
        f"""\
        SELECT
            MIN({time_spine_source.time_column_name})
            , MAX({time_spine_source.time_column_name})
        FROM {time_spine_source.spine_table.sql}
        """,
    )
    assert range_df.shape == (1, 2), f"Expected 1 row with 2 columns in range dataframe, got {range_df}"
    date_range = tuple(range_df.squeeze())

    if sql_client.sql_engine_attributes.timestamp_type_supported:
        assert tuple(map(lambda x: x.strftime(ISO8601_PYTHON_TS_FORMAT), date_range)) == (
            TimeRangeConstraint.ALL_TIME_BEGIN().strftime(ISO8601_PYTHON_TS_FORMAT),
            TimeRangeConstraint.ALL_TIME_END().strftime(ISO8601_PYTHON_TS_FORMAT),
        )

    else:
        assert date_range == (
            TimeRangeConstraint.ALL_TIME_BEGIN().strftime(ISO8601_PYTHON_FORMAT),
            TimeRangeConstraint.ALL_TIME_END().strftime(ISO8601_PYTHON_FORMAT),
        )



import datetime
import logging
import traceback
from typing import List
from metricflow.instances import MaterializationModelReference

from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.source_node import SourceNodeBuilder
from metricflow.dataset.convert_data_source import DataSourceToDataSetConverter
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.dataset.dataset import DataSet
from metricflow.model.objects.materialization import Materialization
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.semantic_model import SemanticModel
from metricflow.model.validations.validator_helpers import (
    FileContext,
    MaterializationContext,
    ModelValidationRule,
    ValidationError,
    ValidationIssueType,
    validate_safely,
    ValidationFutureError,
)
from metricflow.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.query.query_parser import MetricFlowQueryParser

logger = logging.getLogger(__name__)


class ValidMaterializationRule(ModelValidationRule):
    """Check that a materialization config is valid.

    * Metrics listed are valid.
    * Dimensions or identifiers are valid for the listed metrics.
    * Primary time dimension is listed in the dimensions section.
    """

    @staticmethod
    @validate_safely(whats_being_done="checking materialization is defined correctly")
    def _validate_materialization(
        materialization: Materialization,
        mf_query_parser: MetricFlowQueryParser,
    ) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []

        context = MaterializationContext(
            file_context=FileContext.from_metadata(metadata=materialization.metadata),
            materialization=MaterializationModelReference(materialization_name=materialization.name),
        )

        if not materialization.dimensions:
            return [
                ValidationError(
                    context=context,
                    message=f"Materialization '{materialization.name}' does not have dimensions listed",
                )
            ]

        try:
            mf_query_parser.parse_and_validate_query(
                metric_names=materialization.metrics,
                group_by_names=materialization.dimensions,
            )
        # TODO: Using broad exception clause until the query validation returns a list of errors.
        except Exception as err:
            issues.append(
                ValidationFutureError(
                    context=context,
                    message=str(err),
                    error_date=datetime.date(2022, 5, 23),
                    extra_detail="".join(traceback.format_tb(err.__traceback__)),
                )
            )

        # Primary dimension checks
        contained_metric_time_dimension_names: List[str] = []
        for mat_dimension_or_identifier_name in materialization.dimensions:
            structured_spec = StructuredLinkableSpecName.from_name(mat_dimension_or_identifier_name)
            if structured_spec.element_name == DataSet.metric_time_dimension_name():
                contained_metric_time_dimension_names.append(mat_dimension_or_identifier_name)

        if len(contained_metric_time_dimension_names) == 0:
            issues.append(
                # Leaving as a warning to avoid model validation failures during the transition
                ValidationFutureError(
                    context=context,
                    message=f"Metric time dimension {DataSet.metric_time_dimension_name()} not listed"
                    f" as a dimension in materialization {materialization.name}",
                    error_date=datetime.date(2022, 8, 1),
                )
            )

        if len(contained_metric_time_dimension_names) > 1:
            issues.append(
                ValidationError(
                    context=context,
                    message=f"Multiple metric time dimensions {contained_metric_time_dimension_names} listed in "
                    f"materialization {materialization.name}",
                )
            )

        return issues

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring materializations are valid")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        """Check that all of the metrics and dimensions listed in a materialization are valid."""
        issues: List[ValidationIssueType] = []

        semantic_model = SemanticModel(model)
        source_data_sets: List[DataSourceDataSet] = []
        converter = DataSourceToDataSetConverter(
            column_association_resolver=DefaultColumnAssociationResolver(semantic_model)
        )
        for data_source in semantic_model.user_configured_model.data_sources:
            data_set = converter.create_sql_source_data_set(data_source)
            source_data_sets.append(data_set)

        # Any schema will work since we're just using it to render the output.
        time_spine_source = TimeSpineSource(schema_name="dummy_schema")
        node_output_resolver = DataflowPlanNodeOutputDataSetResolver[DataSourceDataSet](
            column_association_resolver=DefaultColumnAssociationResolver(semantic_model),
            semantic_model=semantic_model,
            time_spine_source=time_spine_source,
        )
        source_node_builder = SourceNodeBuilder(semantic_model)
        source_nodes = source_node_builder.create_from_data_sets(source_data_sets)

        mf_query_parser = MetricFlowQueryParser(
            model=SemanticModel(user_configured_model=model),
            source_nodes=source_nodes,
            node_output_resolver=node_output_resolver,
        )

        for materialization in model.materializations:
            issues += ValidMaterializationRule._validate_materialization(
                materialization=materialization,
                mf_query_parser=mf_query_parser,
            )
        return issues



import logging
import threading
from typing import ClassVar, Optional

import pandas as pd
import sqlalchemy
from sqlalchemy.pool import StaticPool

from metricflow.dataflow.sql_table import SqlTable
from metricflow.protocols.sql_client import SupportedSqlEngine, SqlEngineAttributes
from metricflow.sql.render.duckdb_renderer import DuckDbSqlQueryPlanRenderer
from metricflow.sql.render.sql_plan_renderer import SqlQueryPlanRenderer
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.sql_clients.common_client import SqlDialect
from metricflow.sql_clients.sqlalchemy_dialect import SqlAlchemySqlClient

logger = logging.getLogger(__name__)


class DuckDbEngineAttributes(SqlEngineAttributes):
    """Engine-specific attributes for the DuckDb query engine"""

    sql_engine_type: ClassVar[SupportedSqlEngine] = SupportedSqlEngine.DUCKDB

    # SQL Engine capabilities
    date_trunc_supported: ClassVar[bool] = True
    full_outer_joins_supported: ClassVar[bool] = True
    indexes_supported: ClassVar[bool] = True
    multi_threading_supported: ClassVar[bool] = True
    timestamp_type_supported: ClassVar[bool] = True
    timestamp_to_string_comparison_supported: ClassVar[bool] = True
    # Cancelling should be possible, but not yet implemented.
    cancel_submitted_queries_supported: ClassVar[bool] = False

    # SQL Dialect replacement strings
    double_data_type_name: ClassVar[str] = "DOUBLE"
    timestamp_type_name: ClassVar[Optional[str]] = "TIMESTAMP"

    # MetricFlow attributes
    sql_query_plan_renderer: ClassVar[SqlQueryPlanRenderer] = DuckDbSqlQueryPlanRenderer()


class DuckDbSqlClient(SqlAlchemySqlClient):
    """Implements DuckDB."""

    @staticmethod
    def from_connection_details(url: str, password: Optional[str] = None) -> SqlAlchemySqlClient:  # noqa: D
        parsed_url = sqlalchemy.engine.url.make_url(url)
        dialect = SqlDialect.DUCKDB.value
        if parsed_url.drivername != dialect:
            raise ValueError(f"Expected dialect '{dialect}' in {url}")

        if password:
            raise ValueError("Password should be empty")

        return DuckDbSqlClient(file_path=parsed_url.database)

    def __init__(self, file_path: Optional[str] = None) -> None:  # noqa: D
        # DuckDB is not designed with concurrency, but in can work in multi-threaded settings with
        # check_same_thread=False, StaticPool, and serializing of queries via a lock.
        self._concurrency_lock = threading.RLock()
        super().__init__(
            sqlalchemy.create_engine(
                f"duckdb:///{file_path if file_path else ':memory:'}",
                connect_args={"check_same_thread": False},
                poolclass=StaticPool,
            )
        )

    @property
    def sql_engine_attributes(self) -> SqlEngineAttributes:
        """Collection of attributes and features specific to the Snowflake SQL engine"""
        return DuckDbEngineAttributes()

    def cancel_submitted_queries(self) -> None:  # noqa: D
        raise NotImplementedError

    def _engine_specific_query_implementation(  # noqa: D
        self, stmt: str, bind_params: SqlBindParameters
    ) -> pd.DataFrame:
        with self._concurrency_lock:
            return super()._engine_specific_query_implementation(stmt=stmt, bind_params=bind_params)

    def _engine_specific_execute_implementation(self, stmt: str, bind_params: SqlBindParameters) -> None:  # noqa: D
        with self._concurrency_lock:
            return super()._engine_specific_execute_implementation(stmt=stmt, bind_params=bind_params)

    def _engine_specific_dry_run_implementation(self, stmt: str, bind_params: SqlBindParameters) -> None:  # noqa: D
        with self._concurrency_lock:
            return super()._engine_specific_dry_run_implementation(stmt=stmt, bind_params=bind_params)

    def create_table_from_dataframe(  # noqa: D
        self, sql_table: SqlTable, df: pd.DataFrame, chunk_size: Optional[int] = None
    ) -> None:
        with self._concurrency_lock:
            return super().create_table_from_dataframe(
                sql_table=sql_table,
                df=df,
                chunk_size=chunk_size,
            )



import logging
import threading
from collections import OrderedDict
from typing import Set, Union

import pandas as pd
import pytest
from metricflow.dataflow.sql_table import SqlTable
from metricflow.object_utils import assert_values_exhausted, random_id
from metricflow.protocols.sql_client import SqlClient, SupportedSqlEngine
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.sql_clients.sql_utils import make_df
from metricflow.test.compare_df import assert_dataframes_equal
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from sqlalchemy.exc import ProgrammingError

logger = logging.getLogger(__name__)


def _random_table() -> str:
    return f"test_table_{random_id()}"


def _select_x_as_y(sql_client: SqlClient, x: int = 1, y: str = "y") -> str:  # noqa: D
    return f"SELECT {x} AS {y}"


def _check_1col(df: pd.DataFrame, col: str = "y", vals: Set[Union[int, str]] = {1}) -> None:  # noqa: D
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (len(vals), 1)
    assert df.columns.tolist() == [col]
    assert set(df[col]) == vals


def test_query(sql_client: SqlClient) -> None:  # noqa: D
    df = sql_client.query(_select_x_as_y(sql_client))
    _check_1col(df)


def test_query_with_execution_params(sql_client: SqlClient) -> None:  # noqa: D
    expr = "SELECT :x as y"
    sql_execution_params = SqlBindParameters()
    sql_execution_params.param_dict = OrderedDict([("x", "1")])
    df = sql_client.query(expr, sql_bind_parameters=sql_execution_params)
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (1, 1)
    assert df.columns.tolist() == ["y"]
    assert set(df["y"]) == {"1"}


def test_select_one_query(sql_client: SqlClient) -> None:  # noqa: D
    sql_client.query("SELECT 1")
    with pytest.raises(Exception):
        sql_client.query("this is garbage")


def test_failed_query_with_execution_params(sql_client: SqlClient) -> None:  # noqa: D
    expr = "SELECT :x"
    sql_execution_params = SqlBindParameters()
    sql_execution_params.param_dict = OrderedDict([("x", "1")])

    sql_client.query(expr, sql_bind_parameters=sql_execution_params)
    with pytest.raises(Exception):
        sql_client.query("this is garbage")


def test_create_table(mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient) -> None:  # noqa: D
    sql_table = SqlTable(schema_name=mf_test_session_state.mf_source_schema, table_name=_random_table())
    sql_client.create_table_as_select(sql_table, _select_x_as_y(sql_client))
    df = sql_client.query(f"SELECT * FROM {sql_table.sql}")
    _check_1col(df)


def test_create_table_from_dataframe(  # noqa: D
    mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient
) -> None:
    expected_df = make_df(
        sql_client=sql_client,
        columns=["int_col", "str_col", "float_col", "bool_col", "time_col"],
        time_columns={"time_col"},
        data=[
            (1, "abc", 1.23, False, "2020-01-01"),
            (2, "def", 4.56, True, "2020-01-02"),
        ],
    )
    sql_table = SqlTable(schema_name=mf_test_session_state.mf_source_schema, table_name=_random_table())
    sql_client.create_table_from_dataframe(sql_table=sql_table, df=expected_df)

    actual_df = sql_client.query(f"SELECT * FROM {sql_table.sql}")
    assert_dataframes_equal(actual=actual_df, expected=expected_df)


def test_table_exists(mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient) -> None:  # noqa: D
    sql_table = SqlTable(schema_name=mf_test_session_state.mf_source_schema, table_name=_random_table())
    sql_client.create_table_as_select(sql_table, _select_x_as_y(sql_client))
    assert sql_client.table_exists(sql_table)


def test_percent_signs_in_query(sql_client: SqlClient) -> None:  # noqa: D
    stmt = "SELECT foo FROM ( SELECT 'abba' AS foo ) source0 WHERE foo LIKE '%a'"
    sql_client.query(stmt)
    df = sql_client.query(stmt)
    _check_1col(df, col="foo", vals={"abba"})


@pytest.fixture()
def example_df() -> pd.DataFrame:
    """Data frame containing data of different types for testing. DateTime would be good to add."""
    return pd.DataFrame(
        columns=["int_col", "str_col", "float_col", "bool_col"],
        data=[
            (1, "abc", 1.23, False),
            (2, "def", 4.56, True),
        ],
    )


def test_health_checks(mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient) -> None:  # noqa: D
    sql_client.health_checks(schema_name=mf_test_session_state.mf_source_schema)


def test_dry_run(mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient) -> None:  # noqa: D
    test_table = SqlTable(schema_name=mf_test_session_state.mf_source_schema, table_name=_random_table())

    stmt = f"CREATE TABLE {test_table.sql} AS SELECT 1 AS foo"
    sql_client.dry_run(stmt)
    assert not sql_client.table_exists(
        test_table
    ), f"Table {test_table.sql} should not exist if the CREATE TABLE was a dry run."


def test_dry_run_of_bad_query_raises_exception(sql_client: SqlClient) -> None:  # noqa: D
    bad_stmt = "SELECT bad_col"
    # Tests that a bad query raises an exception. Different engines may raise different exceptions e.g.
    # ProgrammingError, OperationalError, google.api_core.exceptions.BadRequest, etc.
    with pytest.raises(Exception, match=r"bad_col"):
        sql_client.dry_run(bad_stmt)


def _issue_sleep_query(sql_client: SqlClient, sleep_time: int) -> None:
    """Issue a query that sleeps for a given number of seconds"""
    engine_type = sql_client.sql_engine_attributes.sql_engine_type
    if engine_type == SupportedSqlEngine.SNOWFLAKE:
        sql_client.execute(f"CALL system$wait({sleep_time}, 'SECONDS')")
    elif engine_type in (
        SupportedSqlEngine.BIGQUERY,
        SupportedSqlEngine.REDSHIFT,
    ):
        raise RuntimeError(f"Sleep yet not supported with {engine_type}")

    assert_values_exhausted(engine_type)


def _supports_sleep_query(sql_client: SqlClient) -> bool:
    """Returns true if the given SQL client is supported by _issue_sleep_query()"""
    engine_type = sql_client.sql_engine_attributes.sql_engine_type
    if engine_type == SupportedSqlEngine.SNOWFLAKE:
        return True
    elif engine_type in (
        SupportedSqlEngine.DUCKDB,
        SupportedSqlEngine.BIGQUERY,
        SupportedSqlEngine.REDSHIFT,
    ):
        return False

    assert_values_exhausted(engine_type)


def test_cancel_submitted_queries(  # noqa: D
    mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient
) -> None:
    if not sql_client.sql_engine_attributes.cancel_submitted_queries_supported:
        pytest.skip(
            f"Cancelling queries not yet supported with {sql_client.sql_engine_attributes.sql_engine_type.name}"
        )

    if not _supports_sleep_query(sql_client):
        pytest.skip(f"Sleep queries not yet supported with {sql_client.sql_engine_attributes.sql_engine_type.name}")

    # Submit a 5s sleep query, but then cancel it after 0.5s.
    def cancel_submitted_queries() -> None:
        try:
            sql_client.cancel_submitted_queries()
        except Exception:
            logger.exception("Got an exception while trying to cancel submitted queries.")

    timer_task = threading.Timer(0.5, cancel_submitted_queries)
    timer_task.start()
    with pytest.raises(ProgrammingError):
        if sql_client.sql_engine_attributes.sql_engine_type == SupportedSqlEngine.SNOWFLAKE:
            _issue_sleep_query(sql_client, 5)



import pytest

from metricflow.model.model_validator import ModelValidator
from metricflow.model.objects.data_source import Mutability, MutabilityType
from metricflow.model.objects.elements.dimension import Dimension, DimensionType, DimensionTypeParams
from metricflow.model.objects.elements.measure import Measure, AggregationType
from metricflow.model.objects.metric import MetricType, MetricTypeParams
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import ModelValidationException
from metricflow.test.model.validations.helpers import data_source_with_guaranteed_meta, metric_with_guaranteed_meta
from metricflow.time.time_granularity import TimeGranularity


@pytest.mark.skip("TODO: Will convert to validation rule")
def test_inconsistent_elements() -> None:  # noqa:D
    dim_name = "ename"
    measure_name = "ename"
    with pytest.raises(ModelValidationException):
        ModelValidator().checked_validations(
            UserConfiguredModel(
                data_sources=[
                    data_source_with_guaranteed_meta(
                        name="s1",
                        sql_query="SELECT foo FROM bar",
                        dimensions=[
                            Dimension(
                                name=dim_name,
                                type=DimensionType.TIME,
                                type_params=DimensionTypeParams(
                                    time_granularity=TimeGranularity.DAY,
                                ),
                            )
                        ],
                        mutability=Mutability(type=MutabilityType.IMMUTABLE),
                    ),
                    data_source_with_guaranteed_meta(
                        name="s2",
                        sql_query="SELECT foo FROM bar",
                        measures=[Measure(name=measure_name, agg=AggregationType.SUM)],
                        mutability=Mutability(type=MutabilityType.IMMUTABLE),
                    ),
                ],
                metrics=[
                    metric_with_guaranteed_meta(
                        name=measure_name,
                        type=MetricType.MEASURE_PROXY,
                        type_params=MetricTypeParams(measures=[measure_name]),
                    )
                ],
                materializations=[],
            )
        )



import logging

import math
import pandas as pd


logger = logging.getLogger(__name__)


def _dataframes_contain_same_data(
    actual: pd.DataFrame,
    expected: pd.DataFrame,
) -> bool:
    """Compare all elements of the dataframe one by one for equality.

    We use math.isclose() to avoid issues with float comparisons.
    """
    if expected.shape != actual.shape:
        return False

    for c in range(expected.shape[0]):
        for r in range(expected.shape[1]):
            # NaNs can't be compared for equality.
            if pd.isna(expected.iloc[c, r]) and pd.isna(actual.iloc[c, r]):
                pass
            elif isinstance(expected.iloc[c, r], float) and isinstance(actual.iloc[c, r], float):
                if not math.isclose(expected.iloc[c, r], actual.iloc[c, r]):
                    return False
            elif expected.iloc[c, r] != actual.iloc[c, r]:
                return False
    return True


def assert_dataframes_equal(
    actual: pd.DataFrame, expected: pd.DataFrame, sort_columns: bool = True, allow_empty: bool = False
) -> None:
    """Check that contents of DataFrames are the same.

    If sort_columns is set to false, value and column order needs to be the same.
    """
    if set(actual.columns) != set(expected.columns):
        raise ValueError(
            f"DataFrames do not contain the same columns. actual: {set(actual.columns)}, "
            f"expected: {set(expected.columns)}"
        )

    if not allow_empty and actual.shape[0] == 0 and expected.shape[0] == 0:
        raise AssertionError("Both dataframes have no rows; likely there is a mistake with the test")

    if sort_columns:
        sort_by = list(sorted(actual.columns.tolist()))
        expected = expected.loc[:, sort_by].sort_values(sort_by).reset_index(drop=True)
        actual = actual.loc[:, sort_by].sort_values(sort_by).reset_index(drop=True)

    if not _dataframes_contain_same_data(actual=actual, expected=expected):
        raise ValueError(
            f"Dataframes not equal.\n"
            f"Expected:\n{expected.to_markdown(index=False)}"
            "\n---\n"
            f"Actual:\n{actual.to_markdown(index=False)}"
        )



import logging

import pytest

from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.semantics.data_source_container import PydanticDataSourceContainer
from metricflow.model.semantics.linkable_spec_resolver import LinkableElementProperties
from metricflow.model.semantics.semantic_containers import DataSourceSemantics, MetricSemantics
from metricflow.specs import MetricSpec, MeasureReference

logger = logging.getLogger(__name__)


@pytest.fixture
def new_data_source_semantics(simple_user_configured_model: UserConfiguredModel) -> DataSourceSemantics:  # Noqa: D
    return DataSourceSemantics(
        model=simple_user_configured_model,
        configured_data_source_container=PydanticDataSourceContainer(simple_user_configured_model.data_sources),
    )


@pytest.fixture
def new_metric_semantics(  # Noqa: D
    simple_user_configured_model: UserConfiguredModel, new_data_source_semantics: DataSourceSemantics
) -> MetricSemantics:
    return MetricSemantics(
        user_configured_model=simple_user_configured_model,
        data_source_semantics=new_data_source_semantics,
    )


def test_get_names(new_data_source_semantics: DataSourceSemantics) -> None:  # noqa: D
    expected = [
        "booking_paid_at",
        "capacity_latest",
        "company_name",
        "country_latest",
        "created_at",
        "ds",
        "ds_partitioned",
        "home_state",
        "home_state_latest",
        "is_instant",
        "is_lux_latest",
        "verification_type",
    ]
    assert sorted([d.element_name for d in new_data_source_semantics.get_dimension_references()]) == expected

    expected = [
        "average_booking_value",
        "bookers",
        "booking_payments",
        "booking_value",
        "bookings",
        "identity_verifications",
        "instant_bookings",
        "largest_listing",
        "listings",
        "max_booking_value",
        "min_booking_value",
        "smallest_listing",
        "txn_revenue",
        "views",
    ]
    assert sorted([m.element_name for m in new_data_source_semantics.measure_references]) == expected

    expected = [
        "company",
        "create_a_cycle_in_the_join_graph",
        "guest",
        "host",
        "listing",
        "lux_listing",
        "user",
        "verification",
    ]
    assert sorted([i.element_name for i in new_data_source_semantics.get_identifier_references()]) == expected


def test_get_elements(new_data_source_semantics: DataSourceSemantics) -> None:  # noqa: D
    for dimension_reference in new_data_source_semantics.get_dimension_references():
        assert (
            new_data_source_semantics.get_dimension(dimension_reference=dimension_reference).reference
            == dimension_reference
        )
    for measure_reference in new_data_source_semantics.measure_references:
        measure_reference = MeasureReference(element_name=measure_reference.element_name)
        assert new_data_source_semantics.get_measure(measure_reference=measure_reference).reference == measure_reference


def test_get_data_sources_for_measure(new_data_source_semantics: DataSourceSemantics) -> None:  # noqa: D
    bookings_sources = new_data_source_semantics.get_data_sources_for_measure(MeasureReference(element_name="bookings"))
    assert len(bookings_sources) == 1
    assert bookings_sources[0].name == "bookings_source"

    views_sources = new_data_source_semantics.get_data_sources_for_measure(MeasureReference(element_name="views"))
    assert len(views_sources) == 1
    assert views_sources[0].name == "views_source"

    listings_sources = new_data_source_semantics.get_data_sources_for_measure(MeasureReference(element_name="listings"))
    assert len(listings_sources) == 1
    assert listings_sources[0].name == "listings_latest"


def test_elements_for_metric(new_metric_semantics: MetricSemantics) -> None:  # noqa: D
    assert set(
        [
            x.qualified_name
            for x in new_metric_semantics.element_specs_for_metrics(
                [MetricSpec(element_name="views")],
                without_any_property=frozenset({LinkableElementProperties.DERIVED_TIME_GRANULARITY}),
            )
        ]
    ) == {
        "create_a_cycle_in_the_join_graph",
        "create_a_cycle_in_the_join_graph__booking_paid_at",
        "create_a_cycle_in_the_join_graph__guest",
        "create_a_cycle_in_the_join_graph__host",
        "create_a_cycle_in_the_join_graph__is_instant",
        "create_a_cycle_in_the_join_graph__listing",
        "create_a_cycle_in_the_join_graph__listing__ds",
        "create_a_cycle_in_the_join_graph__listing__capacity_latest",
        "create_a_cycle_in_the_join_graph__listing__country_latest",
        "create_a_cycle_in_the_join_graph__listing__created_at",
        "create_a_cycle_in_the_join_graph__listing__is_lux_latest",
        "create_a_cycle_in_the_join_graph__listing__user",
        "create_a_cycle_in_the_join_graph__listing__lux_listing",
        "ds",
        "ds_partitioned",
        "listing",
        "listing__capacity_latest",
        "listing__country_latest",
        "listing__created_at",
        "listing__ds",
        "listing__is_lux_latest",
        "listing__lux_listing",
        "listing__user",
        "listing__user__company",
        "listing__user__company_name",
        "listing__user__created_at",
        "listing__user__ds_partitioned",
        "listing__user__home_state",
        "listing__user__home_state_latest",
        "user",
        "user__company",
        "user__company_name",
        "user__created_at",
        "user__ds_partitioned",
        "user__home_state",
        "user__home_state_latest",
    }

    local_specs = new_metric_semantics.element_specs_for_metrics(
        metric_specs=[MetricSpec(element_name="views")],
        with_any_property=frozenset({LinkableElementProperties.LOCAL}),
        without_any_property=frozenset({LinkableElementProperties.DERIVED_TIME_GRANULARITY}),
    )
    assert set([x.qualified_name for x in local_specs]) == {
        "create_a_cycle_in_the_join_graph",
        "ds",
        "ds_partitioned",
        "listing",
        "user",
    }


def test_local_linked_elements_for_metric(new_metric_semantics: MetricSemantics) -> None:  # noqa: D
    result = set(
        [
            x.qualified_name
            for x in new_metric_semantics.element_specs_for_metrics(
                [MetricSpec(element_name="listings")],
                with_any_property=frozenset({LinkableElementProperties.LOCAL_LINKED}),
                without_any_property=frozenset({LinkableElementProperties.DERIVED_TIME_GRANULARITY}),
            )
        ]
    )

    assert result == {
        "listing__created_at",
        "listing__country_latest",
        "listing__is_lux_latest",
        "listing__ds",
        "listing__capacity_latest",
    }



from __future__ import annotations

import enum
import re
from typing import Dict, Tuple, List, Optional
from metricflow.instances import (
    DataSourceElementReference,
    DataSourceReference,
    MaterializationModelReference,
    MetricModelReference,
)

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    DataSourceContext,
    DataSourceElementContext,
    DataSourceElementType,
    FileContext,
    MaterializationContext,
    MetricContext,
    ModelValidationRule,
    ValidationContext,
    ValidationError,
    ValidationIssueType,
    validate_safely,
)
from metricflow.object_utils import assert_values_exhausted
from metricflow.specs import ElementReference
from metricflow.time.time_granularity import TimeGranularity


@enum.unique
class MetricFlowReservedKeywords(enum.Enum):
    """Enumeration of reserved keywords with helper for accessing the reason they are reserved"""

    METRIC_TIME = "metric_time"

    @staticmethod
    def get_reserved_reason(keyword: MetricFlowReservedKeywords) -> str:
        """Get the reason a given keyword is reserved. Guarantees an exhaustive switch"""
        if keyword is MetricFlowReservedKeywords.METRIC_TIME:
            return (
                "Used as the query input for creating time series metrics from measures with "
                "different time dimension names."
            )
        else:
            assert_values_exhausted(keyword)


class UniqueAndValidNameRule(ModelValidationRule):
    """Check that names are unique and valid.

    * Names of elements in data sources are unique / valid within the data source.
    * Names of data sources, dimension sets, metric sets, and materializations in the model are unique / valid.
    """

    NAME_REGEX = re.compile(r"\A[a-z][a-z0-9_]*[a-z0-9]\Z")

    @staticmethod
    def check_valid_name(  # noqa: D
        name: str, context: Optional[ValidationContext] = None
    ) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []

        if not UniqueAndValidNameRule.NAME_REGEX.match(name):
            issues.append(
                ValidationError(
                    context=context,
                    message=f"Invalid name `{name}` - names should only consist of lower case letters, numbers, "
                    f"and underscores. In addition, names should start with a lower case letter, and should not end "
                    f"with an underscore, and they must be at least 2 characters long.",
                )
            )
        if name.upper() in TimeGranularity.list_names():
            issues.append(
                ValidationError(
                    context=context,
                    message=f"Invalid name `{name}` - names cannot match reserved time granularity keywords "
                    f"({TimeGranularity.list_names()})",
                )
            )
        if name.lower() in {reserved_name.value for reserved_name in MetricFlowReservedKeywords}:
            reason = MetricFlowReservedKeywords.get_reserved_reason(MetricFlowReservedKeywords(name.lower()))
            issues.append(
                ValidationError(
                    context=context,
                    message=f"Invalid name `{name}` - this name is reserved by MetricFlow. Reason: {reason}",
                )
            )
        return issues

    @staticmethod
    @validate_safely(whats_being_done="checking data source sub element names are unique")
    def _validate_data_source_elements(data_source: DataSource) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []
        element_info_tuples: List[Tuple[ElementReference, str, ValidationContext]] = []

        if data_source.measures:
            for measure in data_source.measures:
                element_info_tuples.append(
                    (
                        measure.reference,
                        "measure",
                        DataSourceElementContext(
                            file_context=FileContext.from_metadata(metadata=data_source.metadata),
                            data_source_element=DataSourceElementReference(
                                data_source_name=data_source.name, element_name=measure.name
                            ),
                            element_type=DataSourceElementType.MEASURE,
                        ),
                    )
                )
        if data_source.identifiers:
            for identifier in data_source.identifiers:
                element_info_tuples.append(
                    (
                        identifier.reference,
                        "identifier",
                        DataSourceElementContext(
                            file_context=FileContext.from_metadata(metadata=data_source.metadata),
                            data_source_element=DataSourceElementReference(
                                data_source_name=data_source.name, element_name=identifier.name
                            ),
                            element_type=DataSourceElementType.IDENTIFIER,
                        ),
                    )
                )
        if data_source.dimensions:
            for dimension in data_source.dimensions:
                element_info_tuples.append(
                    (
                        dimension.reference,
                        "dimension",
                        DataSourceElementContext(
                            file_context=FileContext.from_metadata(metadata=data_source.metadata),
                            data_source_element=DataSourceElementReference(
                                data_source_name=data_source.name, element_name=dimension.name
                            ),
                            element_type=DataSourceElementType.DIMENSION,
                        ),
                    )
                )
        name_to_type: Dict[ElementReference, str] = {}

        for name, _type, context in element_info_tuples:
            if name in name_to_type:
                issues.append(
                    ValidationError(
                        context=context,
                        message=f"In data source `{data_source.name}`, can't use name `{name.element_name}` for a "
                        f"{_type} when it was already used for a {name_to_type[name]}",
                    )
                )
            else:
                name_to_type[name] = _type

        for name, _, context in element_info_tuples:
            issues += UniqueAndValidNameRule.check_valid_name(name=name.element_name, context=context)

        return issues

    @staticmethod
    @validate_safely(whats_being_done="checking model top level element names are sufficiently unique")
    def _validate_top_level_objects(model: UserConfiguredModel) -> List[ValidationIssueType]:
        """Checks names of objects that are not nested."""
        object_info_tuples = []
        if model.data_sources:
            for data_source in model.data_sources:
                object_info_tuples.append(
                    (
                        data_source.name,
                        "data source",
                        DataSourceContext(
                            file_context=FileContext.from_metadata(metadata=data_source.metadata),
                            data_source=DataSourceReference(data_source_name=data_source.name),
                        ),
                    )
                )
        if model.materializations:
            for materialization in model.materializations:
                object_info_tuples.append(
                    (
                        materialization.name,
                        "materialization",
                        MaterializationContext(
                            file_context=FileContext.from_metadata(metadata=materialization.metadata),
                            materialization=MaterializationModelReference(materialization_name=materialization.name),
                        ),
                    )
                )

        name_to_type: Dict[str, str] = {}

        issues: List[ValidationIssueType] = []

        for name, type_, context in object_info_tuples:
            if name in name_to_type:
                issues.append(
                    ValidationError(
                        context=context,
                        message=f"Can't use name `{name}` for a {type_} when it was already used for a "
                        f"{name_to_type[name]}",
                    )
                )
            else:
                name_to_type[name] = type_

        if model.metrics:
            metric_names = set()
            for metric in model.metrics:
                if metric.name in metric_names:
                    issues.append(
                        ValidationError(
                            context=MetricContext(
                                file_context=FileContext.from_metadata(metadata=metric.metadata),
                                metric=MetricModelReference(metric_name=metric.name),
                            ),
                            message=f"Can't use name `{metric.name}` for a metric when it was already used for a metric",
                        )
                    )
                else:
                    metric_names.add(metric.name)

        for name, _, context in object_info_tuples:
            issues += UniqueAndValidNameRule.check_valid_name(name=name, context=context)

        return issues

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring elements have adequately unique names")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues = []
        issues += UniqueAndValidNameRule._validate_top_level_objects(model=model)

        for data_source in model.data_sources:
            issues += UniqueAndValidNameRule._validate_data_source_elements(data_source=data_source)

        return issues



from dataclasses import dataclass
from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap
from typing import List, Optional, TextIO


@dataclass(frozen=True)
class ConfigKey:
    """Dataclass to represent a yaml key"""

    key: str
    value: str = ""
    comment: Optional[str] = None


class YamlTemplateBuilder:
    """Class to construct and write YAML files."""

    @staticmethod
    def write_yaml(config_keys: List[ConfigKey], file_stream: TextIO) -> None:  # noqa: D
        yaml = YAML()
        file_data = CommentedMap()
        for key in config_keys:
            file_data.insert(0, key.key, key.value, comment=key.comment)
        yaml.dump(file_data, file_stream)



from copy import deepcopy
from typing import Tuple

import pytest
from _pytest.fixtures import FixtureRequest

from metricflow.model.data_warehouse_model_validator import (
    DataWarehouseModelValidator,
    DataWarehouseTaskBuilder,
    DataWarehouseValidationTask,
)
from metricflow.model.model_transformer import ModelTransformer
from metricflow.model.objects.data_source import Mutability, MutabilityType
from metricflow.model.objects.elements.dimension import Dimension, DimensionType
from metricflow.model.objects.elements.identifier import Identifier, IdentifierType
from metricflow.model.objects.elements.measure import AggregationType, Measure
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.protocols.sql_client import SqlClient
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.model.validations.helpers import data_source_with_guaranteed_meta
from metricflow.test.plan_utils import assert_snapshot_text_equal, make_schema_replacement_function


@pytest.fixture(scope="session")
def dw_backed_warehouse_validation_model(
    create_data_warehouse_validation_model_tables: bool,
    data_warehouse_validation_model: UserConfiguredModel,
) -> UserConfiguredModel:
    """Model-generating fixture to ensure the underlying tables are created for querying

    Without an explicit invocation of the create_data_warehouse_validation_model_tables fixture the
    tables used by the data_warehouse_validation_model are not guaranteed to exist. This fixture
    guarantees execution of the underlying create table statements for the model, and can be used in
    any test that executes warehouse validation queries. It is not needed for test cases which simply
    use the model to construct tasks without executing them.
    """
    assert create_data_warehouse_validation_model_tables, "Failed to create DW validation tables!"
    return data_warehouse_validation_model


def test_build_data_source_tasks(
    mf_test_session_state: MetricFlowTestSessionState,
    data_warehouse_validation_model: UserConfiguredModel,
    sql_client: SqlClient,
) -> None:  # noqa:D
    tasks = DataWarehouseTaskBuilder.gen_data_source_tasks(
        model=data_warehouse_validation_model,
        sql_client=sql_client,
        system_schema=mf_test_session_state.mf_system_schema,
    )
    assert len(tasks) == len(data_warehouse_validation_model.data_sources)


def test_task_runner(sql_client: SqlClient, mf_test_session_state: MetricFlowTestSessionState) -> None:  # noqa: D
    dw_validator = DataWarehouseModelValidator(
        sql_client=sql_client, system_schema=mf_test_session_state.mf_system_schema
    )

    def good_query() -> Tuple[str, SqlBindParameters]:
        return ("SELECT 'foo' AS foo", SqlBindParameters())

    tasks = [
        DataWarehouseValidationTask(query_and_params_callable=good_query, error_message="Could not select foo"),
    ]

    issues = dw_validator.run_tasks(tasks=tasks)
    assert len(issues.all_issues) == 0

    def bad_query() -> Tuple[str, SqlBindParameters]:
        return ("SELECT (true) AS col1 FROM doesnt_exist", SqlBindParameters())

    err_msg_bad = "Could not access table 'doesnt_exist' in data warehouse"
    bad_task = DataWarehouseValidationTask(query_and_params_callable=bad_query, error_message=err_msg_bad)

    tasks.append(bad_task)
    issues = dw_validator.run_tasks(tasks=tasks)
    assert len(issues.all_issues) == 1
    assert len(issues.errors) == 1
    assert err_msg_bad in issues.errors[0].message


def test_validate_data_sources(  # noqa: D
    dw_backed_warehouse_validation_model: UserConfiguredModel,
    sql_client: SqlClient,
    mf_test_session_state: MetricFlowTestSessionState,
) -> None:
    model = deepcopy(dw_backed_warehouse_validation_model)

    dw_validator = DataWarehouseModelValidator(
        sql_client=sql_client, system_schema=mf_test_session_state.mf_system_schema
    )

    issues = dw_validator.validate_data_sources(model)
    assert len(issues.all_issues) == 0

    model.data_sources.append(
        data_source_with_guaranteed_meta(
            name="test_data_source2",
            sql_table="doesnt.exist",
            dimensions=[],
            mutability=Mutability(type=MutabilityType.IMMUTABLE),
        )
    )

    issues = dw_validator.validate_data_sources(model)
    assert len(issues.all_issues) == 1
    assert len(issues.errors) == 1
    assert "Unable to access data source `test_data_source2`" in issues.all_issues[0].message


def test_build_dimension_tasks(  # noqa: D
    data_warehouse_validation_model: UserConfiguredModel,
    sql_client: SqlClient,
    mf_test_session_state: MetricFlowTestSessionState,
) -> None:
    tasks = DataWarehouseTaskBuilder.gen_dimension_tasks(
        model=data_warehouse_validation_model,
        sql_client=sql_client,
        system_schema=mf_test_session_state.mf_system_schema,
    )
    # on data source query with all dimensions
    assert len(tasks) == 1
    # 1 categorical dimension task, 1 time dimension task, 4 granularity based time dimension tasks
    assert len(tasks[0].on_fail_subtasks) == 6


def test_validate_dimensions(  # noqa: D
    dw_backed_warehouse_validation_model: UserConfiguredModel,
    sql_client: SqlClient,
    mf_test_session_state: MetricFlowTestSessionState,
) -> None:
    model = deepcopy(dw_backed_warehouse_validation_model)

    dw_validator = DataWarehouseModelValidator(
        sql_client=sql_client, system_schema=mf_test_session_state.mf_system_schema
    )

    issues = dw_validator.validate_dimensions(model)
    assert len(issues.all_issues) == 0

    dimensions = list(model.data_sources[0].dimensions)
    dimensions.append(Dimension(name="doesnt_exist", type=DimensionType.CATEGORICAL))
    model.data_sources[0].dimensions = dimensions

    issues = dw_validator.validate_dimensions(model)
    # One isssue is created for the short circuit query failure, and another is
    # created for the sub task checking the specific dimension
    assert len(issues.all_issues) == 2


def test_build_identifiers_tasks(  # noqa: D
    data_warehouse_validation_model: UserConfiguredModel,
    sql_client: SqlClient,
    mf_test_session_state: MetricFlowTestSessionState,
) -> None:
    tasks = DataWarehouseTaskBuilder.gen_identifier_tasks(
        model=data_warehouse_validation_model,
        sql_client=sql_client,
        system_schema=mf_test_session_state.mf_system_schema,
    )
    assert len(tasks) == 1  # on data source query with all identifiers
    assert len(tasks[0].on_fail_subtasks) == 1  # a sub task for each identifier on the data source


def test_validate_identifiers(  # noqa: D
    dw_backed_warehouse_validation_model: UserConfiguredModel,
    sql_client: SqlClient,
    mf_test_session_state: MetricFlowTestSessionState,
) -> None:
    model = deepcopy(dw_backed_warehouse_validation_model)

    dw_validator = DataWarehouseModelValidator(
        sql_client=sql_client, system_schema=mf_test_session_state.mf_system_schema
    )

    issues = dw_validator.validate_identifiers(model)
    assert len(issues.all_issues) == 0

    identifiers = list(model.data_sources[0].identifiers)
    identifiers.append(Identifier(name="doesnt_exist", type=IdentifierType.UNIQUE))
    model.data_sources[0].identifiers = identifiers

    issues = dw_validator.validate_identifiers(model)
    # One isssue is created for the short circuit query failure, and another is
    # created for the sub task checking the specific identifier
    assert len(issues.all_issues) == 2


def test_build_measure_tasks(  # noqa: D
    data_warehouse_validation_model: UserConfiguredModel,
    sql_client: SqlClient,
    mf_test_session_state: MetricFlowTestSessionState,
) -> None:
    tasks = DataWarehouseTaskBuilder.gen_measure_tasks(
        model=data_warehouse_validation_model,
        sql_client=sql_client,
        system_schema=mf_test_session_state.mf_system_schema,
    )
    assert len(tasks) == 1  # on data source query with all measures
    assert len(tasks[0].on_fail_subtasks) == 1  # a sub task for each measure on the data source


def test_validate_measures(  # noqa: D
    dw_backed_warehouse_validation_model: UserConfiguredModel,
    sql_client: SqlClient,
    mf_test_session_state: MetricFlowTestSessionState,
) -> None:
    model = deepcopy(dw_backed_warehouse_validation_model)

    dw_validator = DataWarehouseModelValidator(
        sql_client=sql_client, system_schema=mf_test_session_state.mf_system_schema
    )

    issues = dw_validator.validate_measures(model)
    assert len(issues.all_issues) == 0

    measures = list(model.data_sources[0].measures)
    measures.append(Measure(name="doesnt_exist", agg=AggregationType.SUM, agg_time_dimension="ds"))
    model.data_sources[0].measures = measures

    issues = dw_validator.validate_measures(model)
    # One isssue is created for the short circuit query failure, and another is
    # created for the sub task checking the specific measure
    assert len(issues.all_issues) == 2


def test_build_metric_tasks(  # noqa: D
    request: FixtureRequest,
    data_warehouse_validation_model: UserConfiguredModel,
    sql_client: SqlClient,
    mf_test_session_state: MetricFlowTestSessionState,
) -> None:
    tasks = DataWarehouseTaskBuilder.gen_metric_tasks(
        model=data_warehouse_validation_model,
        system_schema=mf_test_session_state.mf_system_schema,
        sql_client=sql_client,
    )
    assert len(tasks) == 1
    (query_string, _params) = tasks[0].query_and_params_callable()
    assert_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        group_id="data_warehouse_validation_model",
        snapshot_id="query0",
        snapshot_text=query_string,
        snapshot_file_extension=".sql",
        incomparable_strings_replacement_function=make_schema_replacement_function(
            system_schema=mf_test_session_state.mf_system_schema, source_schema=mf_test_session_state.mf_source_schema
        ),
        additional_sub_directories_for_snapshots=(sql_client.__class__.__name__,),
    )


def test_validate_metrics(  # noqa: D
    dw_backed_warehouse_validation_model: UserConfiguredModel,
    sql_client: SqlClient,
    mf_test_session_state: MetricFlowTestSessionState,
) -> None:
    model = deepcopy(dw_backed_warehouse_validation_model)
    dw_validator = DataWarehouseModelValidator(
        sql_client=sql_client, system_schema=mf_test_session_state.mf_system_schema
    )

    issues = dw_validator.validate_metrics(model)
    assert len(issues.all_issues) == 0

    # Update model to have a new measure which creates a new metric by proxy
    new_measures = list(model.data_sources[0].measures)
    new_measures.append(
        Measure(
            name="count_cats",
            agg=AggregationType.SUM,
            expr="is_cat",  # doesn't exist as column
            create_metric=True,
        )
    )
    model.data_sources[0].measures = new_measures
    model.metrics = []
    model = ModelTransformer.pre_validation_transform_model(model)
    model = ModelTransformer.post_validation_transform_model(model)

    # Validate new metric created by proxy causes an issue (because the column used doesn't exist)
    dw_validator = DataWarehouseModelValidator(
        sql_client=sql_client, system_schema=mf_test_session_state.mf_system_schema
    )
    issues = dw_validator.validate_metrics(model)
    assert len(issues.all_issues) == 1
    assert len(issues.errors) == 1
    assert "Unable to query metric `count_cats`" in issues.errors[0].message



from __future__ import annotations

from typing import List, Optional, Sequence

from metricflow.instances import DataSourceReference
from metricflow.model.objects.common import Metadata
from metricflow.model.objects.elements.dimension import Dimension
from metricflow.model.objects.elements.identifier import Identifier
from metricflow.model.objects.elements.measure import Measure
from metricflow.model.objects.base import ModelWithMetadataParsing, HashableBaseModel
from metricflow.object_utils import ExtendedEnum
from metricflow.specs import LinkableElementReference, MeasureReference


class DataSourceOrigin(ExtendedEnum):
    """Describes how data sources were created

    Impacts determination of validity and duration of storage
    """

    SOURCE = "source"  # "input" data sources
    DERIVED = "derived"  # generated by the semantic layer originating (perhaps indirectly) from sources


class MutabilityType(ExtendedEnum):
    """How data at the physical layer is expected to behave"""

    UNKNOWN = "UNKNOWN"
    IMMUTABLE = "IMMUTABLE"  # never changes
    APPEND_ONLY = "APPEND_ONLY"  # appends along an orderable column
    DS_APPEND_ONLY = "DS_APPEND_ONLY"  # appends along daily column
    FULL_MUTATION = "FULL_MUTATION"  # no guarantees, everything may change


class MutabilityTypeParams(HashableBaseModel):
    """Type params add additional context to mutability"""

    min: Optional[str]
    max: Optional[str]
    update_cron: Optional[str]
    along: Optional[str]


class Mutability(HashableBaseModel):
    """Describes the mutability properties of a data source"""

    type: MutabilityType
    type_params: Optional[MutabilityTypeParams]


class DataSource(HashableBaseModel, ModelWithMetadataParsing):
    """Describes a data source"""

    name: str
    description: Optional[str]
    sql_table: Optional[str]
    sql_query: Optional[str]
    dbt_model: Optional[str]

    identifiers: Sequence[Identifier] = []
    measures: Sequence[Measure] = []
    dimensions: Sequence[Dimension] = []

    mutability: Mutability = Mutability(
        type=MutabilityType.FULL_MUTATION, type_params=MutabilityTypeParams(update_cron="0 0,12 * * *")
    )

    origin: DataSourceOrigin = DataSourceOrigin.SOURCE
    metadata: Optional[Metadata]

    @property
    def identifier_references(self) -> List[LinkableElementReference]:  # noqa: D
        return [i.reference for i in self.identifiers]

    @property
    def dimension_references(self) -> List[LinkableElementReference]:  # noqa: D
        return [i.reference for i in self.dimensions]

    @property
    def measure_references(self) -> List[MeasureReference]:  # noqa: D
        return [i.reference for i in self.measures]

    def get_measure(self, measure_reference: MeasureReference) -> Measure:  # noqa: D
        for measure in self.measures:
            if measure.reference == measure_reference:
                return measure

        raise ValueError(
            f"No dimension with name ({measure_reference.element_name}) in data source with name ({self.name})"
        )

    def get_dimension(self, dimension_reference: LinkableElementReference) -> Dimension:  # noqa: D
        for dim in self.dimensions:
            if dim.reference == dimension_reference:
                return dim

        raise ValueError(f"No dimension with name ({dimension_reference}) in data source with name ({self.name})")

    def get_identifier(self, identifier_reference: LinkableElementReference) -> Identifier:  # noqa: D
        for ident in self.identifiers:
            if ident.reference == identifier_reference:
                return ident

        raise ValueError(f"No identifier with name ({identifier_reference}) in data source with name ({self.name})")

    @property
    def partitions(self) -> List[Dimension]:  # noqa: D
        return [dim for dim in self.dimensions or [] if dim.is_partition]

    @property
    def partition(self) -> Optional[Dimension]:  # noqa: D
        partitions = self.partitions
        if not partitions:
            return None
        if len(partitions) > 1:
            raise ValueError(f"too many partitions for data source {self.name}")
        return partitions[0]

    @property
    def reference(self) -> DataSourceReference:  # noqa: D
        return DataSourceReference(data_source_name=self.name)



import copy
import re

from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.model_validator import ModelValidator
from metricflow.object_utils import flatten_nested_sequence
from metricflow.test.test_utils import find_data_source_with


def test_measures_only_exist_in_one_data_source(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    model = copy.deepcopy(simple_model__pre_transforms)

    build = ModelValidator().validate_model(model)
    duplicate_measure_message = "Found measure with name .* in multiple data sources with names"
    found_issue = False

    if build.issues is not None:
        for issue in build.issues.all_issues:
            if re.search(duplicate_measure_message, issue.message):
                found_issue = True

    assert found_issue is False

    # add measure present in one data source to another
    first_data_source, _ = find_data_source_with(
        model, lambda data_source: data_source.measures is not None and len(data_source.measures) > 0
    )

    second_data_source, _ = find_data_source_with(
        model, lambda data_source: data_source.measures is not None and data_source.name != first_data_source.name
    )

    measure = first_data_source.measures[0]
    second_data_source.measures = list(flatten_nested_sequence([second_data_source.measures, [measure]]))

    build = ModelValidator().validate_model(model)

    if build.issues is not None:
        for issue in build.issues.all_issues:
            if re.search(duplicate_measure_message, issue.message):
                found_issue = True

    assert found_issue is True



import copy
import random
from metricflow.model.objects.elements.identifier import CompositeSubIdentifier
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.reserved_keywords import RESERVED_KEYWORDS, ReservedKeywordsRule
from metricflow.model.validations.validator_helpers import ValidationIssueLevel
from metricflow.test.test_utils import find_data_source_with


def random_keyword() -> str:  # noqa: D
    return random.choice(RESERVED_KEYWORDS)


def copied_model(simple_model__pre_transforms: UserConfiguredModel) -> UserConfiguredModel:  # noqa: D
    return copy.deepcopy(simple_model__pre_transforms)


def test_no_reserved_keywords(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    issues = ReservedKeywordsRule.validate_model(simple_model__pre_transforms)
    assert len(issues) == 0


def test_reserved_keywords_in_sql_table(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    model = copied_model(simple_model__pre_transforms)
    (data_source_with_sql_table, _index) = find_data_source_with(
        model=model, function=lambda data_source: data_source.sql_table is not None
    )
    data_source_with_sql_table.sql_table = f"{random_keyword()}.{random_keyword()}"

    issues = ReservedKeywordsRule.validate_model(model)
    assert len(issues) == 1
    assert issues[0].level == ValidationIssueLevel.ERROR


def test_reserved_keywords_in_dimensions(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    model = copied_model(simple_model__pre_transforms)
    (data_source, _index) = find_data_source_with(
        model=model, function=lambda data_source: len(data_source.dimensions) > 0
    )
    dimension = data_source.dimensions[0]
    dimension.name = random_keyword()

    issues = ReservedKeywordsRule.validate_model(model)
    assert len(issues) == 1
    assert issues[0].level == ValidationIssueLevel.ERROR


def test_reserved_keywords_in_measures(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    model = copied_model(simple_model__pre_transforms)
    (data_source, _index) = find_data_source_with(
        model=model, function=lambda data_source: len(data_source.measures) > 0
    )
    measure = data_source.measures[0]
    measure.name = random_keyword()

    issues = ReservedKeywordsRule.validate_model(model)
    assert len(issues) == 1
    assert issues[0].level == ValidationIssueLevel.ERROR


def test_reserved_keywords_in_identifiers(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    model = copied_model(simple_model__pre_transforms)
    (data_source, _index) = find_data_source_with(
        model=model, function=lambda data_source: len(data_source.identifiers) > 0
    )
    identifier = data_source.identifiers[0]
    identifier.name = random_keyword()
    identifier.identifiers = []

    issues = ReservedKeywordsRule.validate_model(model)
    assert len(issues) == 1
    assert issues[0].level == ValidationIssueLevel.ERROR


def test_reserved_keywords_in_composite_identifiers(  # noqa: D
    simple_model__pre_transforms: UserConfiguredModel,
) -> None:
    model = copied_model(simple_model__pre_transforms)
    (data_source, _index) = find_data_source_with(
        model=model, function=lambda data_source: len(data_source.identifiers) > 0
    )
    identifier = data_source.identifiers[0]
    identifier.identifiers = [
        CompositeSubIdentifier(name=random_keyword()),  # should error
        CompositeSubIdentifier(name=random_keyword()),  # should error
        CompositeSubIdentifier(expr="SELECT TRUE AS col1"),  # shouldn't error
    ]

    issues = ReservedKeywordsRule.validate_model(model)
    assert len(issues) == 2
    assert issues[0].level == ValidationIssueLevel.ERROR
    assert issues[1].level == ValidationIssueLevel.ERROR



import pytest
from _pytest.fixtures import FixtureRequest

from metricflow.dataflow.sql_table import SqlTable
from metricflow.sql.optimizer.table_alias_simplifier import SqlTableAliasSimplifier
from metricflow.sql.render.sql_plan_renderer import DefaultSqlQueryPlanRenderer, SqlQueryPlanRenderer
from metricflow.sql.sql_exprs import (
    SqlColumnReferenceExpression,
    SqlColumnReference,
    SqlComparisonExpression,
    SqlComparison,
)
from metricflow.sql.sql_plan import (
    SqlSelectColumn,
    SqlTableFromClauseNode,
    SqlSelectStatementNode,
    SqlJoinDescription,
    SqlJoinType,
)
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.sql.compare_sql_plan import assert_default_rendered_sql_equal


@pytest.fixture
def sql_plan_renderer() -> SqlQueryPlanRenderer:  # noqa: D
    return DefaultSqlQueryPlanRenderer()


@pytest.fixture
def base_select_statement() -> SqlSelectStatementNode:
    """SELECT statement used to build test cases.

    -- test0
    SELECT
      from_source.col0 AS from_source_col0
      , joined_source.col0 AS joined_source_col0
    FROM (
      -- from_source
      SELECT
        from_source_table.col0 AS col0
        , from_source_table.join_col AS join_col
      FROM demo.from_source_table from_source_table
    ) from_source
    JOIN (
      -- joined_source
      SELECT
        joined_source_table.col0 AS col0
        , joined_source_table.join_col AS join_col
      FROM demo.joined_source_table joined_source_table
    ) joined_source
    ON
      from_source.join_col = joined_source.join_col
    """
    return SqlSelectStatementNode(
        description="test0",
        select_columns=(
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="from_source", column_name="col0")
                ),
                column_alias="from_source_col0",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="joined_source", column_name="col0")
                ),
                column_alias="joined_source_col0",
            ),
        ),
        from_source=SqlSelectStatementNode(
            description="from_source",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(
                            table_alias="from_source_table",
                            column_name="col0",
                        )
                    ),
                    column_alias="col0",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(
                            table_alias="from_source_table",
                            column_name="join_col",
                        )
                    ),
                    column_alias="join_col",
                ),
            ),
            from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="from_source_table")),
            from_source_alias="from_source_table",
            joins_descs=(),
            group_bys=(),
            order_bys=(),
        ),
        from_source_alias="from_source",
        joins_descs=(
            SqlJoinDescription(
                right_source=SqlSelectStatementNode(
                    description="joined_source",
                    select_columns=(
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(
                                    table_alias="joined_source_table",
                                    column_name="col0",
                                )
                            ),
                            column_alias="col0",
                        ),
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(
                                    table_alias="joined_source_table",
                                    column_name="join_col",
                                )
                            ),
                            column_alias="join_col",
                        ),
                    ),
                    from_source=SqlTableFromClauseNode(
                        sql_table=SqlTable(schema_name="demo", table_name="joined_source_table")
                    ),
                    from_source_alias="joined_source_table",
                    joins_descs=(),
                    group_bys=(),
                    order_bys=(),
                ),
                right_source_alias="joined_source",
                on_condition=SqlComparisonExpression(
                    left_expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="from_source", column_name="join_col")
                    ),
                    comparison=SqlComparison.EQUALS,
                    right_expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="joined_source", column_name="join_col")
                    ),
                ),
                join_type=SqlJoinType.INNER,
            ),
        ),
        where=None,
        group_bys=(),
        order_bys=(),
    )


def test_table_alias_simplification(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    base_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests a case where no pruning should occur"""
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=base_select_statement,
        plan_id="before_alias_simplification",
    )
    simplified_select_node = SqlTableAliasSimplifier().optimize(base_select_statement)
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=simplified_select_node,
        plan_id="after_alias_simplification",
    )



from contextlib import contextmanager, ExitStack
from unittest.mock import patch
from typing import Generator
import pytest

from metricflow.dag.id_generation import IdGeneratorRegistry


class IdNumberSpace:
    """Defines the numbering of IDs when setting up tests and test fixtures."""

    # Before each test run.
    TEST_START = 0
    # When setting up the ConsistentIdObjectRepository
    CONSISTENT_ID_REPOSITORY = 10000


@pytest.fixture(autouse=True)
def patch_id_generators() -> Generator[None, None, None]:
    """Patch ID generators with a new one to get repeatability in plan outputs before every test.

    Plan outputs contain IDs, so if the IDs are not consistent from run to run, there will be diffs in the actual vs.
    expected outputs during a test.
    """

    with patch_id_generators_helper(start_value=IdNumberSpace.TEST_START):
        yield None


@contextmanager
def patch_id_generators_helper(start_value: int) -> Generator[None, None, None]:
    """Replace ID generators in IdGeneratorRegistry with one that has the given start value."""
    # Create patch context managers for all ID generators in the registry with introspection magic.
    patch_context_managers = [
        patch.object(IdGeneratorRegistry, "_class_name_to_id_generator", {}),
        patch.object(IdGeneratorRegistry, "DEFAULT_START_VALUE", start_value),
    ]

    # Enter the patch context for the patches above.
    with ExitStack() as stack:
        for patch_context_manager in patch_context_managers:
            stack.enter_context(patch_context_manager)  # type: ignore
        # This will un-patch when done with the test.
        yield None



from typing import Optional, TypeVar

from metricflow.object_utils import ExtendedEnum


class SqlDialect(ExtendedEnum):
    """All SQL dialects that MQL currently supports. Value of enum is used in URLs as the dialect."""

    DUCKDB = "duckdb"
    REDSHIFT = "redshift"
    POSTGRESQL = "postgresql"
    MYSQL = "mysql"
    SNOWFLAKE = "snowflake"
    BIGQUERY = "bigquery"


T = TypeVar("T")


def not_empty(value: Optional[T], component_name: str, url: str) -> T:
    """Helper to check the value is not None - otherwise raise a helpful exception."""
    if not value:
        raise ValueError(f"Missing {component_name} in {url}")
    else:
        return value



from metricflow.dataset.dataset import DataSet
from metricflow.time.time_granularity import TimeGranularity

# Shortcuts for referring to the metric time dimension.
MTD = DataSet.metric_time_dimension_name()
MTD_REFERENCE = DataSet.metric_time_dimension_reference()
MTD_SPEC_DAY = DataSet.metric_time_dimension_spec(TimeGranularity.DAY)
MTD_SPEC_MONTH = DataSet.metric_time_dimension_spec(TimeGranularity.MONTH)
MTD_SPEC_YEAR = DataSet.metric_time_dimension_spec(TimeGranularity.YEAR)



from typing import Tuple, Generic, TypeVar, List, Dict

KT = TypeVar("KT")
VT = TypeVar("VT")


class MultiDiGraph(Generic[KT, VT]):
    """Generic data structure representing a directed graph with multiple edges"""

    def __init__(self) -> None:  # noqa :D
        # first key is the "from" node
        # second key is the "to" node
        self._adjacency_dict: Dict[KT, Dict[KT, List[VT]]] = {}

    def __getitem__(self, key: Tuple[KT, KT]) -> List[VT]:
        """Get all edges between the two nodes represented by this key"""
        return self._adjacency_dict[key[0]][key[1]]

    def __contains__(self, key: Tuple[KT, KT]) -> bool:
        """Graph contains at least one edge between nodes represented by ths key"""
        return key[0] in self._adjacency_dict and key[1] in self._adjacency_dict[key[0]]

    def remove_node(self, key: KT) -> None:
        """Removes a node from the graph and all edges connected to it"""
        # remove edges starting from node
        if key in self._adjacency_dict:
            del self._adjacency_dict[key]

        # remove all edges that terminate at the removed node
        for to_dict in self._adjacency_dict.values():
            for to_key in list(to_dict.keys()):
                if to_key == key:
                    del to_dict[to_key]

    def add_edge(self, from_key: KT, to_key: KT, value: VT) -> None:
        """Add an edge from `from_key` to `to_key`"""
        if from_key not in self._adjacency_dict:
            self._adjacency_dict[from_key] = {}
        if to_key not in self._adjacency_dict[from_key]:
            self._adjacency_dict[from_key][to_key] = []
        self._adjacency_dict[from_key][to_key].append(value)

    def adj(self, from_key: KT) -> Dict[KT, List[VT]]:
        """Adjascency view"""
        if from_key not in self._adjacency_dict:
            return {}
        return self._adjacency_dict[from_key]

    def edges(self, from_key: KT) -> List[VT]:
        """All edges originating from `from_key`"""
        if from_key not in self._adjacency_dict:
            return []
        return list(t for lst in self._adjacency_dict[from_key].values() for t in lst)



"""Functions to convert a SQL query plan into a text representation."""

import textwrap

import jinja2
from jinja2 import StrictUndefined

from metricflow.dag.dag_to_text import MetricFlowDagToText
from metricflow.sql.sql_plan import (
    SqlQueryPlanNode,
    SqlQueryPlan,
)


def sql_query_plan_node_as_text(root_node: SqlQueryPlanNode) -> str:
    """Recursively convert the tree represented by the root node into a string."""

    return MetricFlowDagToText().to_text(root_node)


def sql_query_plan_as_text(sql_query_plan: SqlQueryPlan) -> str:
    """Converts the SQL query plan to a text representation that can be used for tests / debugging.

    The text representation is similar to XML.
    """
    component_from_render_node_as_text = sql_query_plan_node_as_text(sql_query_plan.render_node)

    # Under <DataflowPlan>, render all components.
    return jinja2.Template(
        textwrap.dedent(
            """\
            <{{ node_class }}{%- if not inner_contents %}/>{%- else %}>
                {%- if inner_contents %}
                {{ inner_contents | indent(4) }}
                {%- endif %}
            </{{ node_class }}>
            {%- endif %}
            """
        ),
        undefined=StrictUndefined,
    ).render(
        node_class=sql_query_plan.__class__.__name__,
        inner_contents=component_from_render_node_as_text,
    )



from _pytest.fixtures import FixtureRequest

from metricflow.dataflow.builder.dataflow_plan_builder import DataflowPlanBuilder
from metricflow.dataflow.dataflow_plan import MetricTimeDimensionTransformNode
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlQueryPlanConverter
from metricflow.protocols.sql_client import SqlClient
from metricflow.specs import TimeDimensionReference, MetricFlowQuerySpec, MetricSpec
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.plan_conversion.test_dataflow_to_sql_plan import convert_and_check
from metricflow.test.time.metric_time_dimension import MTD_SPEC_DAY


def test_metric_time_dimension_transform_node_using_primary_time(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    dataflow_to_sql_converter: DataflowToSqlQueryPlanConverter[DataSourceDataSet],
    consistent_id_object_repository: ConsistentIdObjectRepository,
    sql_client: SqlClient,
) -> None:
    """Tests converting a PlotTimeDimensionTransform node using the primary time dimension to SQL."""
    source_node = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    metric_time_dimension_transform_node = MetricTimeDimensionTransformNode(
        parent_node=source_node, aggregation_time_dimension_reference=TimeDimensionReference(element_name="ds")
    )
    convert_and_check(
        request=request,
        mf_test_session_state=mf_test_session_state,
        dataflow_to_sql_converter=dataflow_to_sql_converter,
        sql_client=sql_client,
        node=metric_time_dimension_transform_node,
    )


def test_metric_time_dimension_transform_node_using_non_primary_time(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    dataflow_to_sql_converter: DataflowToSqlQueryPlanConverter[DataSourceDataSet],
    consistent_id_object_repository: ConsistentIdObjectRepository,
    sql_client: SqlClient,
) -> None:
    """Tests converting a PlotTimeDimensionTransform node using a non-primary time dimension to SQL."""
    source_node = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    metric_time_dimension_transform_node = MetricTimeDimensionTransformNode(
        parent_node=source_node,
        aggregation_time_dimension_reference=TimeDimensionReference(element_name="booking_paid_at"),
    )
    convert_and_check(
        request=request,
        mf_test_session_state=mf_test_session_state,
        dataflow_to_sql_converter=dataflow_to_sql_converter,
        sql_client=sql_client,
        node=metric_time_dimension_transform_node,
    )


def test_simple_query_with_metric_time_dimension(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    dataflow_to_sql_converter: DataflowToSqlQueryPlanConverter[DataSourceDataSet],
    consistent_id_object_repository: ConsistentIdObjectRepository,
    sql_client: SqlClient,
    dataflow_plan_builder: DataflowPlanBuilder,
) -> None:
    """Tests building a query that uses measures defined from 2 different time dimensions."""
    dataflow_plan = dataflow_plan_builder.build_plan(
        MetricFlowQuerySpec(
            metric_specs=(
                MetricSpec(element_name="bookings"),
                MetricSpec(element_name="booking_payments"),
            ),
            dimension_specs=(),
            time_dimension_specs=(MTD_SPEC_DAY,),
        )
    )

    convert_and_check(
        request=request,
        mf_test_session_state=mf_test_session_state,
        dataflow_to_sql_converter=dataflow_to_sql_converter,
        sql_client=sql_client,
        node=dataflow_plan.sink_output_node.parent_node,
    )



import logging
import time
from typing import Sequence, Optional

from metricflow.dag.id_generation import EXEC_NODE_NOOP
from metricflow.execution.execution_plan import (
    ExecutionPlanTask,
    TaskExecutionResult,
    TaskExecutionError,
    SqlQuery,
)

logger = logging.getLogger(__name__)


class NoOpExecutionPlanTask(ExecutionPlanTask):
    """A no-op task for testing executors."""

    # Error to return if should_error is set.
    EXAMPLE_ERROR = TaskExecutionError("Expected Error")

    def __init__(self, parent_tasks: Sequence[ExecutionPlanTask] = (), should_error: bool = False) -> None:  # noqa: D
        """Constructor.

        Args:
            parent_tasks: Self-explanatory.
            should_error: if true, return an error in the results.
        """
        self._should_error = should_error
        super().__init__(task_id=self.create_unique_id(), parent_nodes=list(parent_tasks))

    @property
    def description(self) -> str:  # noqa: D
        return "Dummy No-Op"

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return EXEC_NODE_NOOP

    def execute(self) -> TaskExecutionResult:  # noqa: D
        start_time = time.time()
        time.sleep(0.01)
        end_time = time.time()
        return TaskExecutionResult(
            start_time=start_time, end_time=end_time, errors=(self.EXAMPLE_ERROR,) if self._should_error else ()
        )

    @property
    def sql_query(self) -> Optional[SqlQuery]:  # noqa: D
        return None



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional

from metricflow.dataflow.sql_table import SqlTable


@dataclass(frozen=True)
class Metric:
    """Dataclass representation of a Metric."""

    name: str
    dimensions: List[Dimension]


@dataclass(frozen=True)
class Dimension:
    """Dataclass representation of a Dimension."""

    name: str


@dataclass(frozen=True)
class Materialization:
    """Object to represent a Metric."""

    name: str
    metrics: List[str]
    dimensions: List[str]
    destination_table: Optional[SqlTable]



from __future__ import annotations

import logging
from collections import defaultdict
from typing import Tuple, List, Set, Dict

from metricflow.sql.optimizer.sql_query_plan_optimizer import SqlQueryPlanOptimizer
from metricflow.sql.sql_exprs import (
    SqlExpressionTreeLineage,
)
from metricflow.sql.sql_plan import (
    SqlQueryPlanNode,
    SqlQueryPlanNodeVisitor,
    SqlTableFromClauseNode,
    SqlSelectStatementNode,
    SqlJoinDescription,
    SqlSelectQueryFromClauseNode,
    SqlSelectColumn,
)

logger = logging.getLogger(__name__)


class SqlColumnPrunerVisitor(SqlQueryPlanNodeVisitor[SqlQueryPlanNode]):
    """Removes unnecessary columns from SELECT statements in the SQL query plan.

    As the visitor traverses up to the parents, it pushes the list of required columns and rewrites the parent nodes.
    """

    def __init__(
        self,
        required_column_aliases: Set[str],
    ) -> None:
        """Constructor.

        Args:
            required_column_aliases: the columns aliases that should not be pruned from the SELECT statements that this
            visits.
        """

        self._required_column_aliases = required_column_aliases

    def _search_for_expressions(
        self, select_node: SqlSelectStatementNode, pruned_select_columns: Tuple[SqlSelectColumn, ...]
    ) -> SqlExpressionTreeLineage:
        """Returns the expressions used in the immediate select statement.

        i.e. this does not return expressions used in sub-queries. pruned_select_columns needs to be passed in since the
        node may have the select columns pruned.
        """
        all_expr_search_results: List[SqlExpressionTreeLineage] = []

        for select_column in pruned_select_columns:
            all_expr_search_results.append(select_column.expr.lineage)

        for join_description in select_node.join_descs:
            all_expr_search_results.append(join_description.on_condition.lineage)

        for group_by in select_node.group_bys:
            all_expr_search_results.append(group_by.expr.lineage)

        for order_by in select_node.order_bys:
            all_expr_search_results.append(order_by.expr.lineage)

        if select_node.where:
            all_expr_search_results.append(select_node.where.lineage)

        return SqlExpressionTreeLineage.combine(all_expr_search_results)

    def _prune_columns_from_grandparents(
        self, node: SqlSelectStatementNode, pruned_select_columns: Tuple[SqlSelectColumn, ...]
    ) -> SqlSelectStatementNode:
        """Assume that you need all columns from the parent and prune the grandparents."""
        pruned_from_source: SqlQueryPlanNode
        if node.from_source.as_select_node:
            from_visitor = SqlColumnPrunerVisitor(
                required_column_aliases={x.column_alias for x in node.from_source.as_select_node.select_columns}
            )
            pruned_from_source = node.from_source.as_select_node.accept(from_visitor)
        else:
            pruned_from_source = node.from_source
        pruned_join_descriptions: List[SqlJoinDescription] = []
        for join_description in node.join_descs:
            right_source_as_select_node = join_description.right_source.as_select_node
            if right_source_as_select_node:
                right_source_visitor = SqlColumnPrunerVisitor(
                    required_column_aliases={x.column_alias for x in right_source_as_select_node.select_columns}
                )
                pruned_join_descriptions.append(
                    SqlJoinDescription(
                        right_source=join_description.right_source.accept(right_source_visitor),
                        right_source_alias=join_description.right_source_alias,
                        on_condition=join_description.on_condition,
                        join_type=join_description.join_type,
                    )
                )
            else:
                pruned_join_descriptions.append(join_description)

        return SqlSelectStatementNode(
            description=node.description,
            select_columns=pruned_select_columns,
            from_source=pruned_from_source,
            from_source_alias=node.from_source_alias,
            joins_descs=tuple(pruned_join_descriptions),
            group_bys=node.group_bys,
            order_bys=node.order_bys,
            where=node.where,
            limit=node.limit,
        )

    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> SqlQueryPlanNode:  # noqa: D
        # Remove columns that are not needed from this SELECT statement because the parent SELECT statement doesn't
        # need them. However, keep columns that are in group bys because that changes the meaning of the query.
        pruned_select_columns = tuple(
            select_column
            for select_column in node.select_columns
            if select_column.column_alias in self._required_column_aliases or select_column in node.group_bys
        )

        if len(pruned_select_columns) == 0:
            raise RuntimeError("All columns have been pruned - this indicates an bug in the pruner or in the inputs.")

        # Based on the expressions in this select statement, figure out what column aliases are needed in the sources of
        # this query (i.e. tables or sub-queries in the FROM or JOIN clauses).
        exprs_used_in_this_node = self._search_for_expressions(node, pruned_select_columns)

        # If any of the string expressions don't have context on what columns are used in the expression, then it's
        # impossible to know what columns can be pruned from the parent sources. So return a SELECT statement that
        # leaves the parent sources untouched. Columns from the grandparents can be pruned based on the parent node
        # though.
        if any([string_expr.used_columns is None for string_expr in exprs_used_in_this_node.string_exprs]):
            return self._prune_columns_from_grandparents(node, pruned_select_columns)

        # Create a mapping from the source alias to the column aliases needed from the corresponding source.
        source_alias_to_required_column_alias: Dict[str, Set[str]] = defaultdict(set)
        for column_reference_expr in exprs_used_in_this_node.column_reference_exprs:
            column_reference = column_reference_expr.col_ref
            source_alias_to_required_column_alias[column_reference.table_alias].add(column_reference.column_name)

        # For all string columns, assume that they are needed from all sources since we don't have a table alias
        # in SqlStringExpression.used_columns
        for string_expr in exprs_used_in_this_node.string_exprs:
            if string_expr.used_columns:
                for column_alias in string_expr.used_columns:
                    source_alias_to_required_column_alias[node.from_source_alias].add(column_alias)
                    for join_description in node.join_descs:
                        source_alias_to_required_column_alias[join_description.right_source_alias].add(column_alias)
        # Same with unqualified column references.
        for unqualified_column_reference_expr in exprs_used_in_this_node.column_alias_reference_exprs:
            column_alias = unqualified_column_reference_expr.column_alias
            source_alias_to_required_column_alias[node.from_source_alias].add(column_alias)
            for join_description in node.join_descs:
                source_alias_to_required_column_alias[join_description.right_source_alias].add(column_alias)

        # Once we know which column aliases are required from which source aliases, replace the sources with new SELECT
        # statements.
        from_source_pruner = SqlColumnPrunerVisitor(
            required_column_aliases=source_alias_to_required_column_alias[node.from_source_alias]
        )
        pruned_from_source = node.from_source.accept(from_source_pruner)
        pruned_join_descriptions: List[SqlJoinDescription] = []
        for join_description in node.join_descs:
            join_source_pruner = SqlColumnPrunerVisitor(
                required_column_aliases=source_alias_to_required_column_alias[join_description.right_source_alias]
            )
            pruned_join_descriptions.append(
                SqlJoinDescription(
                    right_source=join_description.right_source.accept(join_source_pruner),
                    right_source_alias=join_description.right_source_alias,
                    on_condition=join_description.on_condition,
                    join_type=join_description.join_type,
                )
            )

        return SqlSelectStatementNode(
            description=node.description,
            select_columns=tuple(pruned_select_columns),
            from_source=pruned_from_source,
            from_source_alias=node.from_source_alias,
            joins_descs=tuple(pruned_join_descriptions),
            group_bys=node.group_bys,
            order_bys=node.order_bys,
            where=node.where,
            limit=node.limit,
        )

    def visit_table_from_clause_node(self, node: SqlTableFromClauseNode) -> SqlQueryPlanNode:  # noqa: D
        """This node is effectively a FROM statement inside a SELECT statement node, so pruning cannot apply."""
        return node

    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> SqlQueryPlanNode:  # noqa: D
        """Pruning cannot be done here since this is an arbitrary user-provided SQL query."""
        return node


class SqlColumnPrunerOptimizer(SqlQueryPlanOptimizer):
    """Removes unnecessary columns in the SELECT clauses."""

    def optimize(self, node: SqlQueryPlanNode) -> SqlQueryPlanNode:  # noqa: D

        # Can't prune columns without knowing the structure of the query.
        if not node.as_select_node:
            return node

        pruning_visitor = SqlColumnPrunerVisitor(
            required_column_aliases={x.column_alias for x in node.as_select_node.select_columns}
        )

        return node.accept(pruning_visitor)



import datetime as dt
import logging
import os.path
import pathlib
import traceback
from functools import wraps
from typing import Any, Callable, List, Optional, Tuple

import click
from dateutil.parser import parse

from metricflow.cli.cli_context import CLIContext
from metricflow.configuration.config_builder import ConfigKey
from metricflow.configuration.constants import (
    CONFIG_DWH_CREDS_PATH,
    CONFIG_DWH_DB,
    CONFIG_DWH_DIALECT,
    CONFIG_DWH_HOST,
    CONFIG_DWH_PASSWORD,
    CONFIG_DWH_PORT,
    CONFIG_DWH_PROJECT_ID,
    CONFIG_DWH_SCHEMA,
    CONFIG_DWH_USER,
    CONFIG_DWH_WAREHOUSE,
    CONFIG_EMAIL,
    CONFIG_MODEL_PATH,
)
from metricflow.sql_clients.common_client import SqlDialect

logger = logging.getLogger(__name__)

# MetricFlow config keys
MF_CONFIG_KEYS = (
    ConfigKey(key=CONFIG_EMAIL, comment="Optional"),
    ConfigKey(
        key=CONFIG_MODEL_PATH,
        value=f"{pathlib.Path.home()}/.metricflow/sample_models",
        comment="Path to directory containing defined models (Leave until after DWH setup)",
    ),
    ConfigKey(key=CONFIG_DWH_SCHEMA),
)
# BigQuery config keys
MF_BIGQUERY_KEYS = (
    ConfigKey(
        key=CONFIG_DWH_CREDS_PATH, comment="Provide the path to the BigQuery credential file, ignore to use ADC auth"
    ),
    ConfigKey(
        key=CONFIG_DWH_PROJECT_ID, comment="Provide the GCP Project ID, ignore if using service account credentials"
    ),
    ConfigKey(key=CONFIG_DWH_DIALECT, value=SqlDialect.BIGQUERY.value),
)

# Redshift config keys
MF_POSTGRESQL_KEYS = (
    ConfigKey(key=CONFIG_DWH_DB),
    ConfigKey(key=CONFIG_DWH_PASSWORD, comment="Password associated with the provided user"),
    ConfigKey(key=CONFIG_DWH_USER, comment="Username for the data warehouse"),
    ConfigKey(key=CONFIG_DWH_PORT),
    ConfigKey(key=CONFIG_DWH_HOST, comment="Host name"),
    ConfigKey(key=CONFIG_DWH_DIALECT, value=SqlDialect.POSTGRESQL.value),
)

# Redshift config keys
MF_REDSHIFT_KEYS = (
    ConfigKey(key=CONFIG_DWH_DB),
    ConfigKey(key=CONFIG_DWH_PASSWORD, comment="Password associated with the provided user"),
    ConfigKey(key=CONFIG_DWH_USER, comment="Username for the data warehouse"),
    ConfigKey(key=CONFIG_DWH_PORT),
    ConfigKey(key=CONFIG_DWH_HOST, comment="Host name"),
    ConfigKey(key=CONFIG_DWH_DIALECT, value=SqlDialect.REDSHIFT.value),
)
# Snowflake config keys
MF_SNOWFLAKE_KEYS = (
    ConfigKey(key=CONFIG_DWH_WAREHOUSE, comment="Provide the warehouse to use"),
    ConfigKey(key=CONFIG_DWH_DB),
    ConfigKey(key=CONFIG_DWH_PASSWORD, comment="Password associated with the provided user"),
    ConfigKey(key=CONFIG_DWH_USER, comment="Username for the data warehouse"),
    ConfigKey(key=CONFIG_DWH_HOST, comment="Snowflake account name"),
    ConfigKey(key=CONFIG_DWH_DIALECT, value=SqlDialect.SNOWFLAKE.value),
)


def generate_duckdb_demo_keys(config_dir: str) -> Tuple[ConfigKey, ...]:
    """Generate configuration keys for DuckDB with a file in the config_dir."""
    return (
        ConfigKey(
            key=CONFIG_DWH_DB, value=os.path.join(config_dir, "duck.db"), comment="For DuckDB, this is the data file."
        ),
        ConfigKey(key=CONFIG_DWH_DIALECT, value="duckdb"),
        ConfigKey(key=CONFIG_DWH_SCHEMA, value="mf_demo"),
    )


# Click Options
def query_options(function: Callable) -> Callable:
    """Common options for a query"""
    function = separated_by_comma_option(
        "--order",
        help_msg='Metrics or dimensions to order by ("-" prefix for DESC). For example: --order -ds or --order ds,-revenue',
        required=False,
    )(function)
    function = click.option(
        "--limit",
        type=str,
        help="Limit the number of rows out using an int or leave blank for no limit. For example: --limit 100",
        callback=lambda ctx, param, value: validate_limit(value),
    )(function)
    function = click.option(
        "--where",
        type=str,
        default=None,
        help='SQL-like where statement provided as a string. For example: --where "revenue > 100"',
    )(function)
    function = start_end_time_options(function)
    function = separated_by_comma_option(
        "--dimensions",
        help_msg="Dimensions to group by: syntax is --dimensions ds or for multiple dimensions --dimensions ds,org",
    )(function)
    function = separated_by_comma_option(
        "--metrics",
        help_msg="Metrics to query for: syntax is --metrics bookings or for multiple metrics --metrics bookings,messages",
    )(function)
    return function


def start_end_time_options(function: Callable) -> Callable:
    """Options for start_time and end_time."""
    function = click.option(
        "--start-time",
        type=str,
        default=None,
        help="Optional iso8601 timestamp to constraint the start time of the data (inclusive)",
        callback=lambda ctx, param, value: convert_to_datetime(value),
    )(function)

    function = click.option(
        "--end-time",
        type=str,
        default=None,
        help="Optional iso8601 timestamp to constraint the end time of the data (inclusive)",
        callback=lambda ctx, param, value: convert_to_datetime(value),
    )(function)
    return function


def separated_by_comma_option(option_name: str, help_msg: str = "", required: bool = True) -> Callable:
    """Parse input containing a string separated by commma to a List."""

    def wraps(function: Callable) -> Callable:
        function = click.option(
            option_name,
            required=required,
            help=help_msg,
            callback=lambda ctx, param, value: parse_comma_separated_inputs(value),
        )(function)
        return function

    return wraps


# Parsers/Validators
def convert_to_datetime(datetime_str: Optional[str]) -> Optional[dt.datetime]:
    """Callback to convert string to datetime given as an iso8601 timestamp."""
    if datetime_str is None:
        return None

    try:
        return parse(datetime_str)
    except Exception:
        raise click.BadParameter("must be valid iso8601 timestamp")


def parse_comma_separated_inputs(value: Optional[str]) -> Optional[List[str]]:  # noqa: D
    # If comma exist, explode this into a list and return
    if value is None:
        return None
    if "," in value:
        return [i.strip() for i in value.split(",")]

    # Return a list of the single value
    return [value]


def validate_limit(limit: Optional[str]) -> Optional[int]:
    """Validates and transform limit input."""
    if limit and not limit.isnumeric():
        raise click.BadParameter("limit must be an int. For no limit, do not pass this argument")
    return int(limit) if limit else None


# Misc
def exception_handler(func: Callable[..., Any]) -> Callable[..., Any]:  # type: ignore[misc]
    """Decorator to handle exceptions."""

    @wraps(func)
    def wrapper(*args, **kwargs):  # type: ignore
        try:
            func(*args, **kwargs)
        except Exception as e:
            # This will log to the file handlers registered in the root.
            logging.exception("Got an exception in the exception handler.")
            # Checks if CLIContext has verbose flag set

            if isinstance(args[0], CLIContext):
                cli_context: CLIContext = args[0]
                click.echo(f"\nERROR: {str(e)}\nLog file: {cli_context.config.log_file_path}")
            else:
                if not isinstance(args[0], CLIContext):
                    logger.error(
                        f"Missing {CLIContext.__name__} as the first argument to the function "
                        f"{getattr(func, '__name__', repr(func))}"
                    )
                click.echo(f"\nERROR: {str(e)}")
            if args and hasattr(args[0], "verbose") and args[0].verbose is True:
                click.echo(traceback.format_exc())

            exit(1)

    return wrapper



import logging
import signal
import sys

import click
import datetime as dt
import jinja2
import os
import pandas as pd
import pathlib
import textwrap
import time

from halo import Halo
from importlib.metadata import version as pkg_version
from packaging.version import parse
from typing import Callable, List, Optional
from update_checker import UpdateChecker

from metricflow.cli import PACKAGE_NAME
from metricflow.cli.constants import DEFAULT_RESULT_DECIMAL_PLACES, MAX_LIST_OBJECT_ELEMENTS
from metricflow.cli.cli_context import CLIContext
from metricflow.cli.tutorial import create_sample_data, gen_sample_model_configs, remove_sample_tables
from metricflow.cli.utils import (
    MF_BIGQUERY_KEYS,
    MF_CONFIG_KEYS,
    MF_REDSHIFT_KEYS,
    MF_SNOWFLAKE_KEYS,
    exception_handler,
    query_options,
    separated_by_comma_option,
    start_end_time_options,
    generate_duckdb_demo_keys,
    MF_POSTGRESQL_KEYS,
)
from metricflow.configuration.config_builder import YamlTemplateBuilder
from metricflow.dataflow.dataflow_plan_to_text import dataflow_plan_as_text
from metricflow.engine.metricflow_engine import MetricFlowQueryRequest, MetricFlowExplainResult, MetricFlowQueryResult
from metricflow.engine.utils import model_build_result_from_config, path_to_models
from metricflow.model.data_warehouse_model_validator import DataWarehouseModelValidator
from metricflow.model.model_validator import ModelValidator
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.parsing.config_linter import ConfigLinter
from metricflow.model.validations.validator_helpers import ModelValidationResults
from metricflow.sql_clients.common_client import SqlDialect
from metricflow.telemetry.models import TelemetryLevel
from metricflow.telemetry.reporter import TelemetryReporter, log_call
from metricflow.dag.dag_visualization import display_dag_as_svg

logger = logging.getLogger(__name__)

pass_config = click.make_pass_decorator(CLIContext, ensure=True)
_telemetry_reporter = TelemetryReporter(report_levels_higher_or_equal_to=TelemetryLevel.USAGE)
_telemetry_reporter.add_python_log_handler()
_telemetry_reporter.add_rudderstack_handler()


@click.group()
@click.option("-v", "--verbose", is_flag=True)
@pass_config
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def cli(cfg: CLIContext, verbose: bool) -> None:  # noqa: D
    cfg.verbose = verbose

    checker = UpdateChecker()
    result = checker.check(PACKAGE_NAME, pkg_version(PACKAGE_NAME))
    # result is None when an update was not found or a failure occurred
    if result:
        click.secho(
            "‼️ Warning: A new version of the MetricFlow CLI is available.",
            bold=True,
            fg="red",
        )

        click.echo(
            f"💡 Please update to version {result.available_version}, released {result.release_date} by running:\n"
            f"\t$ pip install --upgrade {PACKAGE_NAME}\n",
        )

    # Cancel queries submitted to the DW if the user precess CTRL + c / process is terminated.
    # Note: docs unclear on the type for the 'frame' argument.
    def exit_signal_handler(signal_type: int, frame) -> None:  # type: ignore
        if signal_type == signal.SIGINT:
            click.echo("Got SIGINT")
        elif signal_type == signal.SIGTERM:
            click.echo("Got SIGTERM")
        else:
            # Shouldn't happen since this should ony be registered for SIGINT / SIGTERM.
            click.echo(f"Got unhandled signal {signal_type}")
            return

        try:
            if cfg.sql_client.sql_engine_attributes.cancel_submitted_queries_supported:
                logger.info("Cancelling submitted queries")
                cfg.sql_client.cancel_submitted_queries()
                cfg.sql_client.close()
        finally:
            sys.exit(-1)

    signal.signal(signal.SIGINT, exit_signal_handler)
    signal.signal(signal.SIGTERM, exit_signal_handler)


@cli.command()
def version() -> None:
    """Print the current version of the MetricFlow CLI."""
    click.echo(pkg_version(PACKAGE_NAME))


@cli.command()
@click.option("--restart", is_flag=True, help="Wipe the config file and start over")
@pass_config
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def setup(cfg: CLIContext, restart: bool) -> None:
    """Setup MetricFlow."""

    click.echo(
        textwrap.dedent(
            """\
            🎉 Welcome to MetricFlow! 🎉
            """
        )
    )

    path = pathlib.Path(cfg.config.file_path)
    abs_path = path.absolute()
    to_create = not path.exists() or restart

    # Seed the config template to the config file
    if to_create:
        dialect_map = {
            SqlDialect.SNOWFLAKE.value: MF_SNOWFLAKE_KEYS,
            SqlDialect.BIGQUERY.value: MF_BIGQUERY_KEYS,
            SqlDialect.REDSHIFT.value: MF_REDSHIFT_KEYS,
            SqlDialect.POSTGRESQL.value: MF_POSTGRESQL_KEYS,
            SqlDialect.DUCKDB.value: generate_duckdb_demo_keys(config_dir=cfg.config.dir_path),
        }

        click.echo("Please enter your data warehouse dialect.")
        click.echo("Use 'duckdb' for a standalone demo.")
        click.echo("")
        dialect = click.prompt(
            "Dialect",
            type=click.Choice(sorted([x for x in dialect_map.keys()])),
            show_choices=True,
        )

        # If there is a collision, prefer to use the key in the dialect.
        config_keys = list(dialect_map[dialect])
        for mf_config_key in MF_CONFIG_KEYS:
            if not any(x.key == mf_config_key.key for x in config_keys):
                config_keys.append(mf_config_key)

        with open(abs_path, "w") as file:
            YamlTemplateBuilder.write_yaml(config_keys, file)

    template_description = (
        f"A template config file has been created in {abs_path}.\n"
        if to_create
        else f"A template config file already exists in {abs_path}, so it was left alone.\n"
    )
    click.echo(
        textwrap.dedent(
            f"""\
            💻 {template_description}
            Next steps:
              1. Review and fill out relevant fields.
              2. Run `mf health-checks` to validate the data warehouse connection.
              3. Run `mf validate-configs` to validate the model configurations.
            """
        )
    )


@cli.command()
@click.option("-m", "--msg", is_flag=True, help="Output the final steps dialogue")
@click.option("--skip-dw", is_flag=True, help="Skip the data warehouse health checks")
@click.option("--drop-tables", is_flag=True, help="Drop all the dummy tables created via tutorial")
@pass_config
@click.pass_context
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def tutorial(ctx: click.core.Context, cfg: CLIContext, msg: bool, skip_dw: bool, drop_tables: bool) -> None:
    """Run user through a tutorial."""

    # This text is also located in the projects top-level README.md
    help_msg = textwrap.dedent(
        """\
        🤓 Please run the following steps,

            1.  In '{$HOME}/.metricflow/config.yml', `model_path` should be '{$HOME}/.metricflow/sample_models'.
            2.  Try validating your data model: `mf validate-configs`
            3.  Check out your metrics: `mf list-metrics`
            4.  Check out dimensions for your metric `mf list-dimensions --metric-names transactions`
            5.  Query your first metric: `mf query --metrics transactions --dimensions ds --order ds`
            6.  Show the SQL MetricFlow generates: `mf query --metrics transactions --dimensions ds --order ds --explain`
            7.  Visualize the plan: `mf query --metrics transactions --dimensions ds --order ds --explain --display-plans`
                * This only works if you have graphviz installed - see README.
                * Aesthetic improvements to the visualization are TBD.
            8.  Add another dimension: `mf query --metrics transactions --dimensions ds,customer__country --order ds`
            9.  Add a higher date granularity: `mf query --metrics transactions --dimensions ds__week --order ds__week`
            10. Try a more complicated query: `mf query --metrics transactions,transaction_usd_na,transaction_usd_na_l7d --dimensions ds,is_large --order ds --start-time 2022-03-20 --end-time 2022-04-01`
                * You can also add `--explain --display-plans`.
            11. For more ways to interact with the sample models, go to ‘https://docs.transform.co/docs/metricflow/metricflow-tutorial’.
            12. Once you’re done, run `mf tutorial --skip-dw --drop-tables` to drop the sample tables.
        """
    )

    if msg:
        click.echo(help_msg)
        exit()

    # Check if the MetricFlow configuration file exists
    path = pathlib.Path(cfg.config.file_path)
    if not path.absolute().exists():
        click.echo("💡 Please run `mf setup` to get your configs set up before going through the tutorial.")
        exit()

    # Validate that the data warehouse connection is successful
    if not skip_dw:
        ctx.invoke(health_checks)
        click.confirm("❓ Are the health-checks all passing? Please fix them before continuing", abort=True)
        click.echo("💡 For future reference, you can continue with the tutorial by adding `--skip-dw`\n")

    if drop_tables:
        spinner = Halo(text="Dropping tables...", spinner="dots")
        spinner.start()
        remove_sample_tables(sql_client=cfg.sql_client, system_schema=cfg.mf_system_schema)
        spinner.succeed("Tables dropped")
        exit()

    # Seed sample data into data warehouse
    spinner = Halo(text=f"🤖 Generating sample data into schema {cfg.mf_system_schema}...", spinner="dots")
    spinner.start()
    created = create_sample_data(sql_client=cfg.sql_client, system_schema=cfg.mf_system_schema)
    if not created:
        spinner.warn("🙊 Skipped creating sample tables since they already exist.")
    else:
        spinner.succeed("📀 Sample tables have been successfully created into your data warehouse.")

    # Seed sample model file
    model_path = os.path.join(cfg.config.dir_path, "sample_models")
    if not os.path.exists(model_path):
        pathlib.Path(model_path).mkdir(parents=True)
    click.echo(f"🤖 Attempting to generate model configs to your local filesystem in '{str(model_path)}'.")
    spinner = Halo(text="Dropping tables...", spinner="dots")
    spinner.start()
    gen_sample_model_configs(dir_path=str(model_path), system_schema=cfg.mf_system_schema)
    spinner.succeed(f"📜 Model configs has been generated into '{model_path}'")

    click.echo(help_msg)
    click.echo("💡 Run `mf tutorial --msg` to see this message again without executing everything else")
    exit()


@cli.command()
@query_options
@click.option(
    "--as-table",
    required=False,
    type=str,
    help="Output the data to a specified SQL table in the form of '<schema>.<table>'",
)
@click.option(
    "--csv",
    type=click.File("wb"),
    required=False,
    help="Provide filepath for dataframe output to csv",
)
@click.option(
    "--explain",
    is_flag=True,
    required=False,
    default=False,
    help="In the query output, show the query that was executed against the data warehouse",
)
@click.option(
    "--display-plans",
    is_flag=True,
    required=False,
    help="Display plans (e.g. metric dataflow) in the browser",
)
@click.option(
    "--decimals",
    required=False,
    default=2,
    help="Choose the number of decimal places to round for the numerical values",
)
@click.option(
    "--show-sql-descriptions",
    is_flag=True,
    default=False,
    help="Shows inline descriptions of nodes in displayed SQL",
)
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def query(
    cfg: CLIContext,
    metrics: List[str],
    dimensions: List[str],
    where: Optional[str] = None,
    start_time: Optional[dt.datetime] = None,
    end_time: Optional[dt.datetime] = None,
    order: Optional[List[str]] = None,
    limit: Optional[int] = None,
    as_table: Optional[str] = None,
    csv: Optional[click.utils.LazyFile] = None,
    explain: bool = False,
    display_plans: bool = False,
    decimals: int = DEFAULT_RESULT_DECIMAL_PLACES,
    show_sql_descriptions: bool = False,
) -> None:
    """Create a new query with MetricFlow and assembles a MetricFlowQueryResult."""
    start = time.time()
    spinner = Halo(text="Initiating query…", spinner="dots")
    spinner.start()

    mf_request = MetricFlowQueryRequest.create_with_random_request_id(
        metric_names=metrics,
        group_by_names=dimensions,
        limit=limit,
        time_constraint_start=start_time,
        time_constraint_end=end_time,
        where_constraint=where,
        order_by_names=order,
        output_table=as_table,
    )

    explain_result: Optional[MetricFlowExplainResult] = None
    query_result: Optional[MetricFlowQueryResult] = None

    if explain:
        explain_result = cfg.mf.explain(mf_request=mf_request)
    else:
        query_result = cfg.mf.query(mf_request=mf_request)

    spinner.succeed(f"Success 🦄 - query completed after {time.time() - start:.2f} seconds")

    if explain:
        assert explain_result
        sql = (
            explain_result.rendered_sql_without_descriptions.sql_query
            if not show_sql_descriptions
            else explain_result.rendered_sql.sql_query
        )
        click.echo("🔎 Generated Dataflow Plan + SQL (remove --explain to see data):")
        click.echo(
            textwrap.indent(
                jinja2.Template(
                    textwrap.dedent(
                        """\
                        Metric Dataflow Plan:
                            {{ plan_text | indent(4) }}
                        """
                    ),
                    undefined=jinja2.StrictUndefined,
                ).render(plan_text=dataflow_plan_as_text(explain_result.dataflow_plan)),
                prefix="-- ",
            )
        )
        click.echo("")
        click.echo(sql)
        if display_plans:
            svg_path = display_dag_as_svg(explain_result.dataflow_plan, cfg.config.dir_path)
            click.echo("")
            click.echo(f"Plan SVG saved to: {svg_path}")
        exit()

    assert query_result
    df = query_result.result_df
    # Show the data if returned successfully
    if df is not None:
        if df.empty:
            click.echo("🕳 Successful MQL query returned an empty result set.")
        elif csv is not None:
            # csv is a LazyFile that is file-like that works in this case.
            df.to_csv(csv, index=False)  # type: ignore
            click.echo(f"🖨 Successfully written query output to {csv.name}")
        else:
            # NOTE: remove `to_string` if no pandas dependency is < 1.1.0
            if parse(pd.__version__) >= parse("1.1.0"):
                click.echo(df.to_markdown(index=False, floatfmt=f".{decimals}f"))
            else:
                click.echo(df.to_string(index=False, float_format=lambda x: format(x, f".{decimals}f")))

        if display_plans:
            svg_path = display_dag_as_svg(query_result.dataflow_plan, cfg.config.dir_path)
            click.echo(f"Plan SVG saved to: {svg_path}")


@cli.command()
@click.option("--search", required=False, type=str, help="Filter available metrics by this search term")
@click.option("--show-all-dims", is_flag=True, default=False, help="Show all dimensions associated with a metric.")
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def list_metrics(cfg: CLIContext, show_all_dims: bool = False, search: Optional[str] = None) -> None:
    """List the metrics with their available dimensions.

    Automatically truncates long lists of dimensions, pass --show-all-dims to see all.
    """

    spinner = Halo(text="🔍 Looking for all available metrics...", spinner="dots")
    spinner.start()

    metrics = cfg.mf.list_metrics()

    if not metrics:
        spinner.fail("List of metrics unavailable.")

    filter_msg = ""
    if search is not None:
        num_metrics = len(metrics)
        metrics = [m for m in metrics if search.lower() in m.name.lower()]
        filter_msg = f" matching `{search}`, of a total of {num_metrics} available"

    spinner.succeed(f"🌱 We've found {len(metrics)} metrics{filter_msg}.")
    click.echo('The list below shows metrics in the format of "metric_name: list of available dimensions"')
    num_dims_to_show = MAX_LIST_OBJECT_ELEMENTS
    for m in metrics:
        # sort dimensions by whether they're local first(if / then global else local) then the dim name
        dimensions = sorted(map(lambda d: d.name, filter(lambda d: "/" not in d.name, m.dimensions))) + sorted(
            map(lambda d: d.name, filter(lambda d: "/" in d.name, m.dimensions))
        )
        if show_all_dims:
            num_dims_to_show = len(dimensions)
        click.echo(
            f"• {click.style(m.name, bold=True, fg='green')}: {', '.join(dimensions[:num_dims_to_show])}"
            + (f" and {len(dimensions) - num_dims_to_show} more" if len(dimensions) > num_dims_to_show else "")
        )


@cli.command()
@separated_by_comma_option(
    "--metric-names", help_msg="List dimensions by given metrics (intersection). Ex. --metric-names bookings,messages"
)
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def list_dimensions(cfg: CLIContext, metric_names: List[str]) -> None:
    """List all unique dimensions."""
    spinner = Halo(
        text="🔍 Looking for all available dimensions...",
        spinner="dots",
    )
    spinner.start()

    dimensions = cfg.mf.simple_dimensions_for_metrics(metric_names)
    if not dimensions:
        spinner.fail("List of dimensions unavailable.")

    spinner.succeed(f"🌱 We've found {len(dimensions)} common dimensions for metrics {metric_names}.")
    for d in dimensions:
        click.echo(f"• {click.style(d.name, bold=True, fg='green')}")


@cli.command()
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def health_checks(cfg: CLIContext) -> None:
    """Performs a health check against the DW provided in the configs."""
    spinner = Halo(
        text="🏥 Running health checks against your data warehouse... (This should not take longer than 30s for a successful connection)",
        spinner="dots",
    )
    spinner.start()
    res = cfg.run_health_checks()
    spinner.succeed("Health checks completed.")
    for test in res:
        test_res = res[test]
        if test_res["status"] != "SUCCESS":
            click.echo(f"• ❌ {click.style(test, bold=True, fg=('red'))}:  Failed with - {test_res['error_message']}.")
        else:
            click.echo(f"• ✅ {click.style(test, bold=True, fg=('green'))}: Success!")


@cli.command()
@click.option("--dimension-name", required=True, type=str, help="Dimension to query values from")
@click.option("--metric-name", required=True, type=str, help="Metric that is associated with the dimension")
@start_end_time_options
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def get_dimension_values(
    cfg: CLIContext,
    metric_name: str,
    dimension_name: str,
    start_time: Optional[dt.datetime] = None,
    end_time: Optional[dt.datetime] = None,
) -> None:
    """List all dimension values with the corresponding metric."""
    spinner = Halo(
        text=f"🔍 Retrieving dimension values for dimension '{dimension_name}' of metric '{metric_name}'...",
        spinner="dots",
    )
    spinner.start()

    dim_vals: Optional[List[str]] = None

    try:
        dim_vals = cfg.mf.get_dimension_values(
            metric_name=metric_name,
            get_group_by_values=dimension_name,
            time_constraint_start=start_time,
            time_constraint_end=end_time,
        )
    except Exception as e:
        spinner.fail()
        click.echo(
            textwrap.dedent(
                f"""\
                ❌ Failed to query dimension values for dimension {dimension_name} of metric {metric_name}.
                    ERROR: {str(e)}
                """
            )
        )
        exit(1)

    assert dim_vals
    spinner.succeed(
        f"🌱 We've found {len(dim_vals)} dimension values for dimension {dimension_name} of metric {metric_name}."
    )
    for dim_val in dim_vals:
        click.echo(f"• {click.style(dim_val, bold=True, fg='green')}")


@cli.command()
@click.option("--search", required=False, type=str, help="Filter available materializations by this search term")
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def list_materializations(cfg: CLIContext, search: Optional[str] = None) -> None:
    """List the materializations with their available metrics and dimensions."""

    spinner = Halo(text="🔍 Looking for all available materializations...", spinner="dots")
    spinner.start()

    materializations = cfg.mf.list_materializations()
    if not materializations:
        spinner.fail("List of materializations unavailable.")

    filter_msg = ""
    if search is not None:
        count = len(materializations)
        materializations = [m for m in materializations if search.lower() in m.name.lower()]
        filter_msg = f" matching `{search}`, of a total of {count} available"

    spinner.succeed(f"🌱 We've found {len(materializations)} materializations{filter_msg}.")
    click.echo(
        'The list below shows materializations in the format of "materialization: details related to materialization"'
    )
    for m in materializations:
        dimensions = sorted(m.dimensions)
        metrics = sorted(m.metrics)
        # Materialization name
        click.echo(f"• {click.style(m.name, bold=True, fg='green')}:")
        # Metrics related to this materalization
        click.echo(
            f"Metrics: {', '.join(metrics[:MAX_LIST_OBJECT_ELEMENTS])}"
            + (
                f" and {len(metrics) - MAX_LIST_OBJECT_ELEMENTS} more"
                if len(metrics) > MAX_LIST_OBJECT_ELEMENTS
                else ""
            )
        )
        # Dimensions related to this materalization
        click.echo(
            f"Dimensions: {', '.join(dimensions[:MAX_LIST_OBJECT_ELEMENTS])}"
            + (
                f" and {len(dimensions) - MAX_LIST_OBJECT_ELEMENTS} more"
                if len(dimensions) > MAX_LIST_OBJECT_ELEMENTS
                else ""
            )
        )
        # Misc related to this materialization
        click.echo(f"destination table: {m.destination_table or m.name}")


@cli.command()
@click.option(
    "--materialization-name",
    required=True,
    type=str,
    help="Name of materialization to materialize",
)
@start_end_time_options
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def materialize(
    cfg: CLIContext,
    materialization_name: str,
    start_time: Optional[dt.datetime] = None,
    end_time: Optional[dt.datetime] = None,
) -> None:
    """Create a new materialization query and returns materialized table"""
    if start_time is None and not click.confirm(
        "You haven't provided a start_time. This means we will materialize from the beginning of time. This may be expensive. Are you sure you want to continue?"
    ):
        click.echo("Exiting")
        exit()

    start = time.time()
    spinner = Halo(text="Initiating materialization query…", spinner="dots")
    spinner.start()

    result_table = cfg.mf.materialize(
        materialization_name,
        time_constraint_start=start_time,
        time_constraint_end=end_time,
    )

    spinner.succeed(f"Success 🦄 - materialize query completed after {time.time() - start:.2f} seconds.")
    click.echo(f"Materialized table created at: {result_table.sql}")


@cli.command()
@click.option(
    "--materialization-name",
    required=True,
    type=str,
    help="Name of materialization to drop",
)
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def drop_materialization(cfg: CLIContext, materialization_name: str) -> None:
    """Drops a given materialized table."""

    start = time.time()
    spinner = Halo(text="Initiating drop materialization query…", spinner="dots")
    spinner.start()

    result = cfg.mf.drop_materialization(materialization_name=materialization_name)

    if result:
        spinner.succeed(f"Success 🦄 - drop materialization query completed after {time.time() - start:.2f} seconds.")
    else:
        spinner.warn(f"Materialized table for `{materialization_name}` did not exist, no table was dropped")


def _print_issues(
    issues: ModelValidationResults, show_non_blocking: bool = False, verbose: bool = False
) -> None:  # noqa: D
    for issue in issues.errors:
        print(f"• {issue.as_cli_formatted_str(verbose=verbose)}")
    if show_non_blocking:
        for issue in issues.future_errors:
            print(f"• {issue.as_cli_formatted_str(verbose=verbose)}")
        for issue in issues.warnings:
            print(f"• {issue.as_cli_formatted_str(verbose=verbose)}")


def _run_dw_validations(
    validation_func: Callable[[UserConfiguredModel, Optional[int]], ModelValidationResults],
    validation_type: str,
    model: UserConfiguredModel,
    timeout: Optional[int],
) -> ModelValidationResults:
    """Helper handles the calling of data warehouse issue generating functions"""

    spinner = Halo(text=f"Validating {validation_type} against data warehouse...", spinner="dots")
    spinner.start()

    results = validation_func(model, timeout)
    if not results.has_blocking_issues:
        spinner.succeed(f"🎉 Successfully validated {validation_type} against data warehouse ({results.summary()})")
    else:
        spinner.fail(
            f"Breaking issues found when validating {validation_type} against data warehouse ({results.summary()})"
        )
    return results


def _data_warehouse_validations_runner(
    dw_validator: DataWarehouseModelValidator, model: UserConfiguredModel, timeout: Optional[int]
) -> ModelValidationResults:
    """Helper which calls the individual data warehouse validations to run and prints collected issues"""

    data_source_results = _run_dw_validations(
        dw_validator.validate_data_sources, model=model, validation_type="data sources", timeout=timeout
    )
    dimension_results = _run_dw_validations(
        dw_validator.validate_dimensions, model=model, validation_type="dimensions", timeout=timeout
    )
    identifier_results = _run_dw_validations(
        dw_validator.validate_identifiers, model=model, validation_type="identifiers", timeout=timeout
    )
    measure_results = _run_dw_validations(
        dw_validator.validate_measures, model=model, validation_type="measures", timeout=timeout
    )
    metric_results = _run_dw_validations(
        dw_validator.validate_metrics, model=model, validation_type="metrics", timeout=timeout
    )

    return ModelValidationResults.merge(
        [data_source_results, dimension_results, identifier_results, measure_results, metric_results]
    )


@cli.command()
@click.option(
    "--dw-timeout", required=False, type=int, help="Optional timeout for data warehouse validation steps. Default None."
)
@click.option(
    "--skip-dw",
    is_flag=True,
    default=False,
    help="If specified, skips the data warehouse validations",
)
@click.option("--show-all", is_flag=True, default=False, help="If specified, prints warnings and future-errors")
@click.option(
    "--verbose-issues", is_flag=True, default=False, help="If specified, prints any extra details issues might have"
)
@click.option(
    "--semantic-validation-workers",
    required=False,
    type=int,
    default=1,
    help="Optional. Uses the number of workers specified to run the semantic validations. Should only be used for exceptionally large configs",
)
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def validate_configs(
    cfg: CLIContext,
    dw_timeout: Optional[int] = None,
    skip_dw: bool = False,
    show_all: bool = False,
    verbose_issues: bool = False,
    semantic_validation_workers: int = 1,
) -> None:
    """Perform validations against the defined model configurations."""
    cfg.verbose = True

    if not show_all:
        print("(To see warnings and future-errors, run again with flag `--show-all`)")

    # Lint Validation
    lint_spinner = Halo(text="Checking for YAML format issues", spinner="dots")
    lint_spinner.start()

    lint_results = ConfigLinter().lint_dir(path_to_models(handler=cfg.config))
    if not lint_results.has_blocking_issues:
        lint_spinner.succeed(f"🎉 Successfully linted config YAML files ({lint_results.summary()})")
    else:
        lint_spinner.fail(f"Breaking issues found in config YAML files ({lint_results.summary()})")
        _print_issues(lint_results, show_non_blocking=show_all, verbose=verbose_issues)
        return

    # Parsing Validation
    parsing_spinner = Halo(text="Building model from configs", spinner="dots")
    parsing_spinner.start()
    parsing_result = model_build_result_from_config(handler=cfg.config, raise_issues_as_exceptions=False)

    if not parsing_result.issues.has_blocking_issues:
        parsing_spinner.succeed(f"🎉 Successfully built model from configs ({parsing_result.issues.summary()})")
    else:
        parsing_spinner.fail(
            f"Breaking issues found when building model from configs ({parsing_result.issues.summary()})"
        )
        _print_issues(parsing_result.issues, show_non_blocking=show_all, verbose=verbose_issues)
        return

    user_model = parsing_result.model

    # Semantic validation
    semantic_spinner = Halo(text="Validating semantics of built model", spinner="dots")
    semantic_spinner.start()
    semantic_result = ModelValidator(max_workers=semantic_validation_workers).validate_model(user_model)

    if not semantic_result.issues.has_blocking_issues:
        semantic_spinner.succeed(
            f"🎉 Successfully validated the semantics of built model ({semantic_result.issues.summary()})"
        )
    else:
        semantic_spinner.fail(
            f"Breaking issues found when checking semantics of built model ({semantic_result.issues.summary()})"
        )
        _print_issues(semantic_result.issues, show_non_blocking=show_all, verbose=verbose_issues)
        return

    dw_results = ModelValidationResults()
    if not skip_dw:
        dw_validator = DataWarehouseModelValidator(sql_client=cfg.sql_client, system_schema=cfg.mf_system_schema)
        dw_results = _data_warehouse_validations_runner(dw_validator=dw_validator, model=user_model, timeout=dw_timeout)

    merged_results = ModelValidationResults.merge(
        [lint_results, parsing_result.issues, semantic_result.issues, dw_results]
    )
    _print_issues(merged_results, show_non_blocking=show_all, verbose=verbose_issues)


if __name__ == "__main__":
    cli()



from __future__ import annotations

from typing import Optional, List, Dict, Any

from metricflow.model.objects.base import HashableBaseModel
from metricflow.object_utils import ExtendedEnum
from metricflow.specs import IdentifierReference, CompositeSubIdentifierReference


class IdentifierType(ExtendedEnum):
    """Defines uniqueness and completeness of identifier"""

    FOREIGN = "foreign"
    PRIMARY = "primary"
    UNIQUE = "unique"

    # for identifiers that are only rendered, not joined on (used when rendering composite identifiers)
    # RENDER_ONLY = "render_only" # DEPRECATED ? we shouldnt need in the new world?


class CompositeSubIdentifier(HashableBaseModel):
    """CompositeSubIdentifiers either describe or reference the identifiers that comprise a composite identifier"""

    name: Optional[str]
    expr: Optional[str]
    ref: Optional[str]

    @property
    def reference(self) -> CompositeSubIdentifierReference:  # noqa: D
        assert self.name, f"The element name should have been set during model transformation. Got {self}"
        return CompositeSubIdentifierReference(element_name=self.name)


class Identifier(HashableBaseModel):
    """Describes a identifier"""

    name: str
    description: Optional[str]
    type: IdentifierType
    role: Optional[str]
    entity: Optional[str]
    identifiers: List[CompositeSubIdentifier] = []
    expr: Optional[str] = None

    def __init__(  # type: ignore
        self,
        name: str,
        type: IdentifierType,
        role: Optional[str] = None,
        entity: Optional[str] = None,
        identifiers: Optional[List[CompositeSubIdentifier]] = None,
        expr: Optional[str] = None,
        **kwargs: Dict[str, Any],  # the parser may instantiate objects with additional fields (eg __parsing_context__)
    ) -> None:
        """Normal pydantic initializer except we set entity to name"""

        identifiers = identifiers or []

        super().__init__(
            name=name,
            type=type,
            role=role,
            entity=entity,
            identifiers=identifiers,
            expr=expr,
        )
        if self.entity is None:
            self.entity = self.reference.element_name

    @property
    def is_primary_time(self) -> bool:  # noqa: D
        return False

    @property
    def is_composite(self) -> bool:  # noqa: D
        return self.identifiers is not None and len(self.identifiers) > 0

    @property
    def reference(self) -> IdentifierReference:  # noqa: D
        return IdentifierReference(element_name=self.name)



from jsonschema import RefResolver

from metricflow.model.parsing.schema_validator import SchemaValidator

TRANSFORM_OBJECT_NAME_PATTERN = "(?!.*__).*^[a-z][a-z0-9_]*[a-z0-9]$"


# Enums
materialization_location_tableau_values = ["TABLEAU", "tableau"]
materialization_location_values = ["DW", "FAST_CACHE"]
materialization_location_values += [x.lower() for x in materialization_location_values]
materialization_location_values += materialization_location_tableau_values

materialization_format_values = ["WIDE"]
materialization_format_values += [x.lower() for x in materialization_format_values]

metric_types_enum_values = ["MEASURE_PROXY", "RATIO", "EXPR", "CUMULATIVE"]
metric_types_enum_values += [x.lower() for x in metric_types_enum_values]

mutability_type_values = ["IMMUTABLE", "APPEND_ONLY", "FULL_MUTATION", "DS_APPEND_ONLY"]
mutability_type_values += [x.lower() for x in mutability_type_values]

identifier_type_enum_values = ["PRIMARY", "UNIQUE", "FOREIGN", "RENDER_ONLY"]
identifier_type_enum_values += [x.lower() for x in identifier_type_enum_values]

aggregation_type_values = ["SUM", "MIN", "MAX", "AVERAGE", "COUNT_DISTINCT", "BOOLEAN", "SUM_BOOLEAN"]
aggregation_type_values += [x.lower() for x in aggregation_type_values]

time_granularity_values = ["DAY", "WEEK", "MONTH", "QUARTER", "YEAR"]
time_granularity_values += [x.lower() for x in time_granularity_values]

dimension_type_values = ["CATEGORICAL", "TIME"]
dimension_type_values += [x.lower() for x in dimension_type_values]

time_dimension_type_values = ["TIME", "time"]

tableau_params = {
    "$id": "materialization_params_tableau",
    "type": "object",
    "properties": {
        "projects": {
            "type": "array",
            "projects": {"type": "string"},
        },
    },
}

materialization_destination_schema = {
    "$id": "materialization_destination_schema",
    "type": "object",
    "properties": {
        "location": {"enum": materialization_location_values},
        "format": {"enum": materialization_format_values},
        "rollups": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "string"},
            },
        },
        "tableau_params": tableau_params,
    },
    "anyOf": [{"not": {"$ref": "#/definitions/is-tableau"}}, {"required": ["tableau_params"]}],
    "definitions": {"is-tableau": {"properties": {"location": {"enum": materialization_location_tableau_values}}}},
    "additionalProperties": False,
}

metric_input_measure_schema = {
    "$id": "metric_input_measure_schema",
    "oneOf": [
        {"type": "string"},
        {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
            },
            "additionalProperties": False,
        },
    ],
}

metric_type_params_schema = {
    "$id": "metric_type_params",
    "type": "object",
    "properties": {
        "numerator": {"$ref": "metric_input_measure_schema"},
        "denominator": {"$ref": "metric_input_measure_schema"},
        "measure": {"$ref": "metric_input_measure_schema"},
        "measures": {
            "type": "array",
            "items": {"$ref": "metric_input_measure_schema"},
        },
        "expr": {"type": ["string", "boolean"]},
        "window": {"type": "string"},
        "grain_to_date": {"type": "string"},
    },
    "additionalProperties": False,
}

composite_sub_identifier_schema = {
    "$id": "composite_sub_identifier_schema",
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "expr": {"type": ["string", "boolean"]},
        "ref": {"type": "string"},
    },
    "additionalProperties": False,
}

identifier_schema = {
    "$id": "identifier_schema",
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "pattern": TRANSFORM_OBJECT_NAME_PATTERN,
        },
        "type": {"enum": identifier_type_enum_values},
        "role": {"type": "string"},
        "expr": {"type": ["string", "boolean"]},
        "entity": {"type": "string"},
        "identifiers": {
            "type": "array",
            "items": {"$ref": "composite_sub_identifier_schema"},
        },
    },
    "additionalProperties": False,
    "required": ["name", "type"],
}

dimension_type_params_schema = {
    "$id": "dimension_type_params_schema",
    "type": "object",
    "properties": {
        "is_primary": {"type": "boolean"},
        "time_format": {"type": "string"},
        "time_granularity": {"enum": time_granularity_values},
    },
    "additionalProperties": False,
    "required": ["time_granularity"],
}

measure_schema = {
    "$id": "measure_schema",
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "pattern": TRANSFORM_OBJECT_NAME_PATTERN,
        },
        "agg": {"enum": aggregation_type_values},
        "agg_time_dimension": {
            "type": "string",
            "pattern": TRANSFORM_OBJECT_NAME_PATTERN,
        },
        "expr": {"type": ["string", "integer", "boolean"]},
        "create_metric": {"type": "boolean"},
        "create_metric_display_name": {"type": "string"},
    },
    "additionalProperties": False,
    "required": ["name", "agg"],
}

dimension_schema = {
    "$id": "dimension_schema",
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "pattern": TRANSFORM_OBJECT_NAME_PATTERN,
        },
        "type": {"enum": dimension_type_values},
        "is_partition": {"type": "boolean"},
        "expr": {"type": ["string", "boolean"]},
        "type_params": {"$ref": "dimension_type_params_schema"},
    },
    # dimension must have type_params if its a time dimension
    "anyOf": [{"not": {"$ref": "#/definitions/is-time-dimension"}}, {"required": ["type_params"]}],
    "definitions": {
        "is-time-dimension": {
            "properties": {"type": {"enum": time_dimension_type_values}},
            "required": ["type"],
        },
    },
    "additionalProperties": False,
    "required": ["name", "type"],
}

mutability_type_params_schema = {
    "$id": "mutability_type_params_schema",
    "type": "object",
    "properties": {
        "min": {"type": "string"},
        "max": {"type": "string"},
        "update_cron": {"type": "string"},
        "along": {"type": "string"},
    },
    "additionalProperties": False,
}

mutability_schema = {
    "$id": "mutability_schema",
    "type": "object",
    "properties": {
        "type": {"enum": mutability_type_values},
        "type_params": {"$ref": "mutability_type_params_schema"},
    },
    "additionalProperties": False,
    "required": ["type"],
}


# Top level object schemas
metric_schema = {
    "$id": "metric",
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "pattern": TRANSFORM_OBJECT_NAME_PATTERN,
        },
        "type": {"enum": metric_types_enum_values},
        "type_params": {"$ref": "metric_type_params"},
        "constraint": {"type": "string"},
        "where_constraint": {"type": "string"},
    },
    "additionalProperties": False,
    "required": ["name", "type", "type_params"],
}

data_source_schema = {
    "$id": "data_source",
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "pattern": TRANSFORM_OBJECT_NAME_PATTERN,
        },
        "sql_table": {"type": "string"},
        "sql_query": {"type": "string"},
        "dbt_model": {"type": "string"},
        "identifiers": {"type": "array", "items": {"$ref": "identifier_schema"}},
        "measures": {"type": "array", "items": {"$ref": "measure_schema"}},
        "dimensions": {"type": "array", "items": {"$ref": "dimension_schema"}},
        "mutability": {"$ref": "mutability_schema"},
        "constraint": {"$ref": "constraint_schema"},
    },
    "additionalProperties": False,
    "required": ["name"],
}

derived_group_by_element_schema = {
    "$id": "derived_group_by_element_schema",
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "pattern": TRANSFORM_OBJECT_NAME_PATTERN,
        },
        "expr": {"type": ["string", "boolean"]},
        "expr_elements": {
            "type": "array",
            "items": {"type": "string"},
        },
    },
    "additionalProperties": False,
    "required": ["name", "expr"],
}

materialization_schema = {
    "$id": "materialization_schema",
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "pattern": TRANSFORM_OBJECT_NAME_PATTERN,
        },
        "dimensions": {
            "type": "array",
            "items": {"type": "string"},
            "minItems": 1,
        },
        "metrics": {
            "type": "array",
            "items": {"type": "string"},
            "minItems": 1,
        },
        "destinations": {
            "type": "array",
            "items": {"$ref": "materialization_destination_schema"},
        },
        "destination_table": {"type": "string"},
    },
    "additionalProperties": False,
    "required": ["name", "dimensions", "metrics"],
}

schema_store = {
    # Top level schemas
    metric_schema["$id"]: metric_schema,
    data_source_schema["$id"]: data_source_schema,
    derived_group_by_element_schema["$id"]: derived_group_by_element_schema,
    materialization_schema["$id"]: materialization_schema,
    # Sub-object schemas
    metric_input_measure_schema["$id"]: metric_input_measure_schema,
    metric_type_params_schema["$id"]: metric_type_params_schema,
    identifier_schema["$id"]: identifier_schema,
    measure_schema["$id"]: measure_schema,
    dimension_schema["$id"]: dimension_schema,
    dimension_type_params_schema["$id"]: dimension_type_params_schema,
    mutability_schema["$id"]: mutability_schema,
    mutability_type_params_schema["$id"]: mutability_type_params_schema,
    composite_sub_identifier_schema["$id"]: composite_sub_identifier_schema,
    materialization_destination_schema["$id"]: materialization_destination_schema,
}


resolver = RefResolver.from_schema(schema=metric_schema, store=schema_store)
data_source_validator = SchemaValidator(data_source_schema, resolver=resolver)
derived_group_by_element_validator = SchemaValidator(derived_group_by_element_schema, resolver=resolver)
materialization_validator = SchemaValidator(materialization_schema, resolver=resolver)
metric_validator = SchemaValidator(metric_schema, resolver=resolver)



from metricflow.sql.render.expr_renderer import (
    DefaultSqlExpressionRenderer,
    SqlExpressionRenderer,
    SqlExpressionRenderResult,
)
from metricflow.sql.render.sql_plan_renderer import DefaultSqlQueryPlanRenderer
from metricflow.sql.sql_exprs import (
    SqlTimeDeltaExpression,
)
from metricflow.time.time_granularity import TimeGranularity


class DuckDbSqlExpressionRenderer(DefaultSqlExpressionRenderer):
    """Expression renderer for the DuckDB engine."""

    def visit_time_delta_expr(self, node: SqlTimeDeltaExpression) -> SqlExpressionRenderResult:  # noqa: D
        arg_rendered = node.arg.accept(self)
        if node.grain_to_date:
            return SqlExpressionRenderResult(
                sql=f"DATE_TRUNC('{node.granularity.value}', {arg_rendered.sql}::timestamp)",
                execution_parameters=arg_rendered.execution_parameters,
            )

        count = node.count
        granularity = node.granularity
        if granularity == TimeGranularity.QUARTER:
            granularity = TimeGranularity.MONTH
            count *= 3

        return SqlExpressionRenderResult(
            sql=f"{arg_rendered.sql} - INTERVAL {count} {granularity.value}",
            execution_parameters=arg_rendered.execution_parameters,
        )


class DuckDbSqlQueryPlanRenderer(DefaultSqlQueryPlanRenderer):
    """Plan renderer for the DuckDB engine."""

    EXPR_RENDERER = DuckDbSqlExpressionRenderer()

    @property
    def expr_renderer(self) -> SqlExpressionRenderer:  # noqa :D
        return self.EXPR_RENDERER



import logging
from abc import ABC, abstractmethod
from datetime import datetime
from typing import List, Optional, Sequence, Dict, Any

from rudder_analytics.client import Client as RudderstackClient

from metricflow.telemetry.models import FunctionStartEvent, FunctionEndEvent, TelemetryPayload

PayloadType = Dict[Any, Any]  # type: ignore
logger = logging.getLogger(__name__)


class TelemetryHandler(ABC):
    """Base class to record telemetry to some destination."""

    @abstractmethod
    def _write_log(self, client_id: str, payload: PayloadType) -> None:
        """Subclasses should implement this to log a serialized version of the payload."""
        pass

    def log(
        self,
        client_id: str,
        function_start_event: Optional[FunctionStartEvent] = None,
        function_end_event: Optional[FunctionEndEvent] = None,
    ) -> bool:
        """Log an event to telemetry."""
        payload = TelemetryPayload(
            client_id=client_id,
            function_start_events=(function_start_event,) if function_start_event else (),
            function_end_events=(function_end_event,) if function_end_event else (),
        )
        self._write_log(client_id, payload.dict())
        return True


class RudderstackTelemetryHandler(TelemetryHandler):
    """A telemetry client that logs data to Rudderstack."""

    RUDDERSTACK_EVENT_NAME = "MF_OSS"

    # Key for logging to Transform
    TFD_WRITE_KEY: str = "2777X6I2uHwXmCM9hB6GPKcegdn"
    TFD_DATA_PLANE: str = "https://transformdprul.dataplane.rudderstack.com"

    def __init__(  # noqa: D
        self,
        data_plane_url: str = TFD_DATA_PLANE,
        write_key: str = TFD_WRITE_KEY,
    ) -> None:
        self._rudderstack_client = RudderstackClient(
            write_key=write_key, host=data_plane_url, debug=False, on_error=None, send=True, sync_mode=False
        )

    def _write_log(self, client_id: str, payload: PayloadType) -> None:  # noqa: D
        """Write log to rudderstack.

        Added context to handle the error, but seems odd.

        >       msg['context']['traits']['anonymousId'] = msg['anonymousId']
        E       KeyError: 'traits'

        /usr/local/lib/python3.8/site-packages/rudder_analytics/client.py:243: KeyError
        """
        context: PayloadType = {"traits": {}}
        self._rudderstack_client.track(
            anonymous_id=client_id,
            event=RudderstackTelemetryHandler.RUDDERSTACK_EVENT_NAME,
            timestamp=datetime.now(),
            properties=payload,
            context=context,
        )


class ToMemoryTelemetryHandler(TelemetryHandler):
    """Records telemetry events to memory for testing purposes."""

    def __init__(self) -> None:  # noqa: D
        self._payloads: List[TelemetryPayload] = []

    def _write_log(self, client_id: str, payload: PayloadType) -> None:  # noqa: D
        pass

    @property
    def payloads(self) -> Sequence[TelemetryPayload]:  # noqa: D
        return self._payloads

    def log(
        self,
        client_id: str,
        function_start_event: Optional[FunctionStartEvent] = None,
        function_end_event: Optional[FunctionEndEvent] = None,
    ) -> bool:
        """Log an event to telemetry"""
        payload = TelemetryPayload(
            client_id=client_id,
            function_start_events=(function_start_event,) if function_start_event else (),
            function_end_events=(function_end_event,) if function_end_event else (),
        )
        if len(self._payloads) > 10:
            self._payloads.pop()

        self._payloads.append(payload)
        return True



import pytest

from metricflow.model.model_validator import ModelValidator
from metricflow.model.objects.data_source import Mutability, MutabilityType, DataSource
from metricflow.model.objects.elements.dimension import Dimension, DimensionType, DimensionTypeParams
from metricflow.model.objects.elements.measure import Measure, AggregationType
from metricflow.model.objects.metric import MetricType, MetricTypeParams, Metric
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import ModelValidationException
from metricflow.specs import DimensionReference, MeasureReference, TimeDimensionReference
from metricflow.test.model.validations.helpers import data_source_with_guaranteed_meta, metric_with_guaranteed_meta
from metricflow.time.time_granularity import TimeGranularity


def test_incompatible_dimension_type() -> None:  # noqa:D
    with pytest.raises(ModelValidationException, match=r"type conflict for dimension"):
        dim_name = "dim"
        measure_name = "measure"
        model_validator = ModelValidator()
        model_validator.checked_validations(
            UserConfiguredModel(
                data_sources=[
                    data_source_with_guaranteed_meta(
                        name="dim1",
                        sql_query=f"SELECT {dim_name}, {measure_name} FROM bar",
                        measures=[Measure(name=measure_name, agg=AggregationType.SUM)],
                        dimensions=[
                            Dimension(
                                name=dim_name,
                                type=DimensionType.TIME,
                                type_params=DimensionTypeParams(
                                    is_primary=True,
                                    time_granularity=TimeGranularity.DAY,
                                ),
                            )
                        ],
                        mutability=Mutability(type=MutabilityType.IMMUTABLE),
                    ),
                    data_source_with_guaranteed_meta(
                        name="categoricaldim",
                        sql_query="SELECT foo FROM bar",
                        dimensions=[Dimension(name=dim_name, type=DimensionType.CATEGORICAL)],
                        mutability=Mutability(type=MutabilityType.IMMUTABLE),
                    ),
                ],
                metrics=[
                    metric_with_guaranteed_meta(
                        name=measure_name,
                        type=MetricType.MEASURE_PROXY,
                        type_params=MetricTypeParams(measures=[measure_name]),
                    )
                ],
                materializations=[],
            )
        )


def test_incompatible_dimension_is_partition() -> None:  # noqa:D
    with pytest.raises(ModelValidationException, match=r"conflicting is_partition attribute for dimension"):
        dim_name = "dim1"
        measure_name = "measure"
        model_validator = ModelValidator()
        model_validator.checked_validations(
            UserConfiguredModel(
                data_sources=[
                    data_source_with_guaranteed_meta(
                        name="dim1",
                        sql_query=f"SELECT {dim_name}, {measure_name} FROM bar",
                        measures=[Measure(name=measure_name, agg=AggregationType.SUM)],
                        dimensions=[
                            Dimension(
                                name=dim_name,
                                type=DimensionType.TIME,
                                is_partition=True,
                                type_params=DimensionTypeParams(
                                    is_primary=True,
                                    time_granularity=TimeGranularity.DAY,
                                ),
                            )
                        ],
                        mutability=Mutability(type=MutabilityType.IMMUTABLE),
                    ),
                    data_source_with_guaranteed_meta(
                        name="dim2",
                        sql_query="SELECT foo1 FROM bar",
                        dimensions=[
                            Dimension(
                                name=dim_name,
                                type=DimensionType.TIME,
                                is_partition=False,
                                type_params=DimensionTypeParams(
                                    time_granularity=TimeGranularity.DAY,
                                ),
                            )
                        ],
                        mutability=Mutability(type=MutabilityType.IMMUTABLE),
                    ),
                ],
                metrics=[
                    metric_with_guaranteed_meta(
                        name=measure_name,
                        type=MetricType.MEASURE_PROXY,
                        type_params=MetricTypeParams(measures=[measure_name]),
                    )
                ],
                materializations=[],
            )
        )


def test_multiple_primary_time_dimensions() -> None:  # noqa:D
    with pytest.raises(ModelValidationException, match=r"one of many defined as primary"):
        dimension_reference = TimeDimensionReference(element_name="ds")
        dimension_reference2 = DimensionReference(element_name="not_ds")
        measure_reference = MeasureReference(element_name="measure")
        model_validator = ModelValidator()
        model_validator.checked_validations(
            model=UserConfiguredModel(
                data_sources=[
                    DataSource(
                        name="dim1",
                        sql_query=f"SELECT ds, {measure_reference.element_name} FROM bar",
                        measures=[
                            Measure(
                                name=measure_reference.element_name,
                                agg=AggregationType.SUM,
                                agg_time_dimension=dimension_reference.element_name,
                            )
                        ],
                        dimensions=[
                            Dimension(
                                name=dimension_reference.element_name,
                                type=DimensionType.TIME,
                                type_params=DimensionTypeParams(
                                    is_primary=True,
                                    time_granularity=TimeGranularity.DAY,
                                ),
                            ),
                            Dimension(
                                name=dimension_reference2.element_name,
                                type=DimensionType.TIME,
                                type_params=DimensionTypeParams(
                                    is_primary=True,
                                    time_granularity=TimeGranularity.DAY,
                                ),
                            ),
                        ],
                        mutability=Mutability(type=MutabilityType.IMMUTABLE),
                    ),
                ],
                metrics=[
                    Metric(
                        name=measure_reference.element_name,
                        type=MetricType.MEASURE_PROXY,
                        type_params=MetricTypeParams(measures=[measure_reference.element_name]),
                    )
                ],
                materializations=[],
            )
        )



from metricflow.time.time_granularity import TimeGranularity

# Python formatting string to use for converting datetime to ISO8601
ISO8601_PYTHON_FORMAT = "%Y-%m-%d"
ISO8601_PYTHON_TS_FORMAT = "%Y-%m-%d %H:%M:%S"

SUPPORTED_GRANULARITIES = [
    TimeGranularity.DAY,
    TimeGranularity.WEEK,
    TimeGranularity.MONTH,
    TimeGranularity.QUARTER,
    TimeGranularity.YEAR,
]
